// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"

#include <stdlib.h>
#include <unistd.h>

#include "utils/logger.h"
#include "utils/memory.h"

#ifndef RUNSTATEDIR
#define RUNSTATEDIR "/var/run/lora"
#endif

struct config *config_init(void)
{
	struct config *config = NULL;

	if ((config = zalloc(sizeof(*config))) == NULL) {
		eerror("Failed to allocate configuration");
		return NULL;

	} else if ((config->core = config_core_init()) == NULL) {
		error("Failed to initialize core configuration");
		return NULL;
	}

	return config;
}

static int parse_api(struct core *core, JSON_Object *root)
{
	const char *string = NULL;
	JSON_Array *array = NULL;
	size_t nb;

	if ((array = json_object_dotget_array(root, "api.uplink")) != NULL) {
		for (nb = 0; nb < json_array_get_count(array); nb++) {
			if ((string = json_array_get_string(array, nb)) == NULL) {
				error("Invalid JSON file: Endpoint is not a string");
				return -1;

			} else if (stringlist_stradd(&core->config->api_uplinks, string) == NULL) {
				eerror("Failed to allocate endpoint");
				return -1;

			} else {
				info("Using api.uplink: %s", string);
			}
		}
	}

	if ((array = json_object_dotget_array(root, "api.downlink")) != NULL) {
		for (nb = 0; nb < json_array_get_count(array); nb++) {
			if ((string = json_array_get_string(array, nb)) == NULL) {
				error("Invalid JSON file: Endpoint is not a string");
				return -1;

			} else if (stringlist_stradd(&core->config->api_downlinks, string) == NULL) {
				eerror("Failed to allocate endpoint");
				return -1;

			} else {
				info("Using api.downlink: %s", string);
			}
		}
	}

	if ((array = json_object_dotget_array(root, "api.control")) != NULL) {
		for (nb = 0; nb < json_array_get_count(array); nb++) {
			if ((string = json_array_get_string(array, nb)) == NULL) {
				error("Invalid JSON file: Endpoint is not a string");
				return -1;

			} else if (stringlist_stradd(&core->config->api_controls, string) == NULL) {
				eerror("Failed to allocate endpoint");
				return -1;

			} else {
				info("Using api.control: %s", string);
			}
		}
	}

	return 0;
}

int config_parse_file(struct core *core, const char *filename)
{
	JSON_Value *root = NULL;

	if (access(filename, R_OK) < 0) {
		eerror("Cannot read file");
		return -1;

	} else if ((root = json_parse_file_with_comments(filename)) == NULL) {
		error("Invalid JSON file");
		return -1;

	} else if (config_core_parse(core, json_value_get_object(root)) < 0) {
		error("Failed to parse core configuration");
		json_value_free(root);
		return -1;

	} else if (parse_api(core, json_value_get_object(root)) < 0) {
		error("Failed to parse API configuration");
		json_value_free(root);
		return -1;
	}

	json_value_free(root);
	return 0;
}

int config_apply(struct core *core)
{
	if (config_core_apply(core) < 0) {
		error("Failed to apply core configuration");
		return -1;

	} else if (core->config->api_uplinks.nb_strings == 0 &&
		   stringlist_stradd(&core->config->api_uplinks,
				     "ipc://" RUNSTATEDIR "/uplink") == NULL) {
		eerror("Failed to allocate endpoint");
		return -1;

	} else if (core->config->api_downlinks.nb_strings == 0 &&
		   stringlist_stradd(&core->config->api_downlinks,
				     "ipc://" RUNSTATEDIR "/downlink") == NULL) {
		eerror("Failed to allocate endpoint");
		return -1;

	} else if (core->config->api_controls.nb_strings == 0 &&
		   stringlist_stradd(&core->config->api_controls,
				     "ipc://" RUNSTATEDIR "/lorad") == NULL) {
		eerror("Failed to allocate endpoint");
		return -1;
	}

	return 0;
}

void config_free(struct core *core)
{
	if (core && core->config) {
		config_core_free(core);
		stringlist_free(&core->config->api_uplinks);
		stringlist_free(&core->config->api_downlinks);
		stringlist_free(&core->config->api_controls);
		free(core->config);
	}
}
