// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils/logger.h"
#include "utils/memory.h"

#define JSON_PARSE_BOOLEAN(object, name, value)                                                            \
	{                                                                                                  \
		JSON_Value *_value = NULL;                                                                 \
		if ((_value = json_object_dotget_value(object, name)) != NULL) {                           \
			if (json_value_get_type(_value) == JSONBoolean) {                                  \
				value = json_value_get_boolean(_value);                                    \
				debug("Value \"" name "\" configured to: %s", (value) ? "true" : "false"); \
			} else {                                                                           \
				warning("Wrong data type for " name);                                      \
			}                                                                                  \
		}                                                                                          \
	}

#define JSON_PARSE_NUMBER(object, name, value)                                                \
	{                                                                                     \
		JSON_Value *_value = NULL;                                                    \
		if ((_value = json_object_dotget_value(object, name)) != NULL) {              \
			if (json_value_get_type(_value) == JSONNumber) {                      \
				value = json_value_get_number(_value);                        \
				debug("Value \"" name "\" configured to: %g", (double)value); \
			} else {                                                              \
				warning("Wrong data type for " name);                         \
			}                                                                     \
		}                                                                             \
	}

#define JSON_PARSE_STRING(object, name, value)                                        \
	{                                                                             \
		JSON_Value *_value = NULL;                                            \
		if ((_value = json_object_dotget_value(object, name)) != NULL) {      \
			if (json_value_get_type(_value) == JSONString) {              \
				value = json_value_get_string(_value);                \
				debug("Value \"" name "\" configured to: %s", value); \
			} else {                                                      \
				warning("Wrong data type for " name);                 \
			}                                                             \
		}                                                                     \
	}

static int parse_sx1301_lbt(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t i;

	if ((object = json_object_get_object(root, "lbt_cfg")) != NULL) {
		JSON_PARSE_BOOLEAN(object, "enable", config_core->lbt.enable);
		JSON_PARSE_NUMBER(object, "rssi_target", config_core->lbt.rssi_target);
		JSON_PARSE_NUMBER(object, "sx127x_rssi_offset", config_core->lbt.rssi_offset);

		if ((array = json_object_get_array(object, "chan_cfg")) != NULL) {
			for (i = 0; i < json_array_get_count(array) && i < LBT_CHANNEL_FREQ_NB; i++) {
				if ((object = json_array_get_object(array, i)) != NULL) {
					config_core->lbt.nb_channel = i + 1;
					JSON_PARSE_NUMBER(object, "freq_hz", config_core->lbt.channels[i].freq_hz);
					JSON_PARSE_NUMBER(object, "scan_time_us", config_core->lbt.channels[i].scan_time_us);
				}
			}
		}
	}

	return 0;
}

static int parse_sx1301_lut(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	char name[32];
	int i;

	for (i = 0; i < TX_GAIN_LUT_SIZE_MAX; i++) {
		snprintf(name, sizeof(name), "tx_lut_%u", i);

		if ((object = json_object_get_object(root, name)) != NULL) {
			config_core->txlut.size = i + 1;
			JSON_PARSE_NUMBER(object, "dig_gain", config_core->txlut.lut[i].dig_gain);
			JSON_PARSE_NUMBER(object, "pa_gain", config_core->txlut.lut[i].pa_gain);
			JSON_PARSE_NUMBER(object, "dac_gain", config_core->txlut.lut[i].dac_gain);
			JSON_PARSE_NUMBER(object, "mix_gain", config_core->txlut.lut[i].mix_gain);
			JSON_PARSE_NUMBER(object, "rf_power", config_core->txlut.lut[i].rf_power);
		}
	}

	return 0;
}

static int parse_sx1301_rf(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	const char *str = NULL;
	char name[32];
	int i;

	for (i = 0; i < LGW_RF_CHAIN_NB; i++, str = NULL) {
		snprintf(name, sizeof(name), "radio_%u", i);

		if ((object = json_object_get_object(root, name)) != NULL) {
			JSON_PARSE_BOOLEAN(object, "enable", config_core->rxrf[i].enable);
			JSON_PARSE_NUMBER(object, "freq", config_core->rxrf[i].freq_hz);
			JSON_PARSE_NUMBER(object, "rssi_offset", config_core->rxrf[i].rssi_offset);
			JSON_PARSE_BOOLEAN(object, "tx_enable", config_core->rxrf[i].tx_enable);
			JSON_PARSE_NUMBER(object, "tx_freq_min", config_core->rfchains[i].tx_freq_min);
			JSON_PARSE_NUMBER(object, "tx_freq_max", config_core->rfchains[i].tx_freq_max);
			JSON_PARSE_NUMBER(object, "tx_notch_freq", config_core->rxrf[i].tx_notch_freq);
			JSON_PARSE_STRING(object, "type", str);

			if (str != NULL && strncmp(str, "SX1255", 6) == 0) {
				config_core->rxrf[i].type = LGW_RADIO_TYPE_SX1255;
			} else if (str != NULL && strncmp(str, "SX1257", 6) == 0) {
				config_core->rxrf[i].type = LGW_RADIO_TYPE_SX1257;
			}
		}
	}

	return 0;
}

static int parse_sx1301_if(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	uint32_t fdev = 0;
	uint32_t bw = 0;
	uint32_t sf = 0;
	char name[32];
	int i;

	/* Set configuration for Lora multi-SF channels */
	for (i = 0; i < LGW_MULTI_NB; ++i) {
		snprintf(name, sizeof(name), "chan_multiSF_%u", i);

		if ((object = json_object_get_object(root, name)) != NULL) {
			JSON_PARSE_BOOLEAN(object, "enable", config_core->rxif[i].enable);
			JSON_PARSE_NUMBER(object, "radio", config_core->rxif[i].rf_chain);
			JSON_PARSE_NUMBER(object, "if", config_core->rxif[i].freq_hz);
		}
	}

	/* Set configuration for Lora standard channel */
	if ((object = json_object_get_object(root, "chan_Lora_std")) != NULL) {
		JSON_PARSE_BOOLEAN(object, "enable", config_core->rxif[8].enable);
		JSON_PARSE_NUMBER(object, "radio", config_core->rxif[8].rf_chain);
		JSON_PARSE_NUMBER(object, "if", config_core->rxif[8].freq_hz);
		JSON_PARSE_NUMBER(object, "bandwidth", bw);
		JSON_PARSE_NUMBER(object, "spread_factor", sf);

		switch (bw) {
		case 500000:
			config_core->rxif[8].bandwidth = BW_500KHZ;
			break;
		case 250000:
			config_core->rxif[8].bandwidth = BW_250KHZ;
			break;
		case 125000:
			config_core->rxif[8].bandwidth = BW_125KHZ;
			break;
		default:
			config_core->rxif[8].bandwidth = BW_UNDEFINED;
			break;
		}

		switch (sf) {
		case 7:
			config_core->rxif[8].datarate = DR_LORA_SF7;
			break;
		case 8:
			config_core->rxif[8].datarate = DR_LORA_SF8;
			break;
		case 9:
			config_core->rxif[8].datarate = DR_LORA_SF9;
			break;
		case 10:
			config_core->rxif[8].datarate = DR_LORA_SF10;
			break;
		case 11:
			config_core->rxif[8].datarate = DR_LORA_SF11;
			break;
		case 12:
			config_core->rxif[8].datarate = DR_LORA_SF12;
			break;
		default:
			config_core->rxif[8].datarate = DR_UNDEFINED;
			break;
		}
	}

	/* Set configuration for FSK channel */
	if ((object = json_object_get_object(root, "chan_FSK")) != NULL) {
		JSON_PARSE_BOOLEAN(object, "enable", config_core->rxif[9].enable);
		JSON_PARSE_NUMBER(object, "radio", config_core->rxif[9].rf_chain);
		JSON_PARSE_NUMBER(object, "if", config_core->rxif[9].freq_hz);
		JSON_PARSE_NUMBER(object, "datarate", config_core->rxif[9].datarate);
		JSON_PARSE_NUMBER(object, "bandwidth", bw);
		JSON_PARSE_NUMBER(object, "freq_deviation", fdev);

		/* if "bandwidth" is set, it has priority over "freq_deviation" */
		if ((bw == 0) && (fdev != 0)) {
			bw = 2 * fdev + config_core->rxif[9].datarate;
		}

		if (bw == 0) {
			config_core->rxif[9].bandwidth = BW_UNDEFINED;
		} else if (bw <= 7800) {
			config_core->rxif[9].bandwidth = BW_7K8HZ;
		} else if (bw <= 15600) {
			config_core->rxif[9].bandwidth = BW_15K6HZ;
		} else if (bw <= 31200) {
			config_core->rxif[9].bandwidth = BW_31K2HZ;
		} else if (bw <= 62500) {
			config_core->rxif[9].bandwidth = BW_62K5HZ;
		} else if (bw <= 125000) {
			config_core->rxif[9].bandwidth = BW_125KHZ;
		} else if (bw <= 250000) {
			config_core->rxif[9].bandwidth = BW_250KHZ;
		} else if (bw <= 500000) {
			config_core->rxif[9].bandwidth = BW_500KHZ;
		} else {
			config_core->rxif[9].bandwidth = BW_UNDEFINED;
		}
	}

	return 0;
}

static int parse_sx1301(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;

	if ((object = json_object_get_object(root, "SX1301_conf")) == NULL) {
		info("No SX1301 configuration");
		return 0;
	}

	/* Board configuration */
	JSON_PARSE_BOOLEAN(object, "lorawan_public", config_core->board.lorawan_public);
	JSON_PARSE_NUMBER(object, "clksrc", config_core->board.clksrc);

	/* Gateway specificities */
	JSON_PARSE_NUMBER(object, "antenna_gain", config_core->antenna_gain);
	JSON_PARSE_NUMBER(object, "insertion_loss", config_core->insertion_loss);

	if (parse_sx1301_lbt(core, object) < 0) {
		error("Failed to parse LBT configuration");
		return -1;

	} else if (parse_sx1301_lut(core, object) < 0) {
		error("Failed to parse LUT configuration");
		return -1;

	} else if (parse_sx1301_rf(core, object) < 0) {
		error("Failed to parse RF configuration");
		return -1;

	} else if (parse_sx1301_if(core, object) < 0) {
		error("Failed to parse IF configuration");
		return -1;
	}

	return 0;
}

static int parse_gateway(struct core *core, JSON_Object *root)
{
	struct config_beacon *beacon = &core->config->core->beacon;
	JSON_Object *object = NULL;
	int sf = 0, bw = 0;

	if ((object = json_object_get_object(root, "gateway_conf")) == NULL) {
		info("No gateway configuration");
		return 0;
	}

	JSON_PARSE_BOOLEAN(object, "beacon_enable", beacon->enable);
	JSON_PARSE_NUMBER(object, "beacon_period", beacon->period);
	JSON_PARSE_NUMBER(object, "beacon_freq_hz", beacon->frequency);
	JSON_PARSE_NUMBER(object, "beacon_freq_nb", beacon->frequencies);
	JSON_PARSE_NUMBER(object, "beacon_freq_step", beacon->frequency_step);
	JSON_PARSE_NUMBER(object, "beacon_datarate", sf);
	JSON_PARSE_NUMBER(object, "beacon_bw_hz", bw);
	JSON_PARSE_NUMBER(object, "beacon_power", beacon->power);
	JSON_PARSE_NUMBER(object, "beacon_infodesc", beacon->infodesc);
	JSON_PARSE_NUMBER(object, "ref_latitude", beacon->latitude);
	JSON_PARSE_NUMBER(object, "ref_longitude", beacon->longitude);

	if (bw) {
		switch (bw) {
		case 125000:
			beacon->bandwidth = BW_125KHZ;
			break;
		case 500000:
			beacon->bandwidth = BW_500KHZ;
			break;
		default:
			error("Invalid beacon bandwidth: %i", bw);
			return -1;
		}
	}

	if (sf) {
		switch (sf) {
		case 8:
			beacon->spreading_factor = DR_LORA_SF8;
			break;
		case 9:
			beacon->spreading_factor = DR_LORA_SF9;
			break;
		case 10:
			beacon->spreading_factor = DR_LORA_SF10;
			break;
		case 12:
			beacon->spreading_factor = DR_LORA_SF12;
			break;
		default:
			error("Invalid beacon spreading factor: %i", sf);
			return -1;
		}
	}

	if (beacon->period > 0 && beacon->period < 6) {
		error("Invalid configuration for Beacon period, must be >= 6s");
		return -1;
	}

	return 0;
}

struct config_core *config_core_init(void)
{
	struct config_core *config_core = NULL;
	size_t n;

	if ((config_core = zalloc(sizeof(*config_core))) == NULL) {
		eerror("Failed to allocate core configuration");
		return NULL;
	}

	for (n = 0; n < TX_GAIN_LUT_SIZE_MAX; n++) {
		config_core->txlut.lut[n].dac_gain = 3;
	}

	config_core->beacon.frequency = 869525000;
	config_core->beacon.frequencies = 1;
	config_core->beacon.bandwidth = BW_125KHZ;
	config_core->beacon.spreading_factor = DR_LORA_SF9;
	config_core->beacon.power = 14;

	return config_core;
}

int config_core_parse(struct core *core, JSON_Object *root)
{
	if (parse_sx1301(core, root) < 0) {
		error("Failed to parse SX1301 configuration");
		return -1;

	} else if (parse_gateway(core, root) < 0) {
		error("Failed to parse gateway configuration");
		return -1;
	}

	return 0;
}

int config_core_apply(struct core *core)
{
	struct config_core *config_core = core->config->core;
	int i;

	if (lgw_board_setconf(config_core->board) != LGW_HAL_SUCCESS) {
		error("Failed to configure board");
		return -1;

	} else if (lgw_lbt_setconf(config_core->lbt) != LGW_HAL_SUCCESS) {
		error("Failed to configure LBT");
		return -1;

	} else if (lgw_txgain_setconf(&(config_core->txlut)) != LGW_HAL_SUCCESS) {
		error("Failed to configure concentrator TX Gain LUT");
		return -1;
	}

	for (i = 0; i < LGW_RF_CHAIN_NB; i++) {
		/* Compute RSSI offset taking into account antenna gain and insertion loss */
		config_core->rxrf[i].rssi_offset = config_core->rxrf[i].rssi_offset -
						   config_core->antenna_gain + config_core->insertion_loss;
		info("rf_chain %d: new RSSI offset including antenna gain (%f) and path loss (%f): %.1f\n",
		     i, config_core->antenna_gain, config_core->insertion_loss, config_core->rxrf[i].rssi_offset);

		if (lgw_rxrf_setconf(i, config_core->rxrf[i]) != LGW_HAL_SUCCESS) {
			error("Failed to configure radio %u", i);
			return -1;
		}
	}

	for (i = 0; i < LGW_IF_CHAIN_NB; i++) {
		if (lgw_rxif_setconf(i, config_core->rxif[i]) != LGW_HAL_SUCCESS) {
			error("Failed to configure channel %u", i);
			return -1;
		}
	}

	return 0;
}

void config_core_free(struct core *core)
{
	if (core && core->config) {
		free(core->config->core);
	}
}
