// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "gpio.h"

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "utils/logger.h"
#include "utils/memory.h"

struct gpio {
	int id;
	int direction;
	int value;
};

#define GPIO_SYS_EXPORT "/sys/class/gpio/export"
#define GPIO_SYS_UNEXPORT "/sys/class/gpio/unexport"
#define GPIO_SYS_ROOT "/sys/class/gpio/gpio%u"
#define GPIO_SYS_DIRECTION GPIO_SYS_ROOT "/direction"
#define GPIO_SYS_VALUE GPIO_SYS_ROOT "/value"

static int gpio_is_exported(int id)
{
	char *path = NULL;
	int ret;

	if (aprintf(&path, GPIO_SYS_ROOT, id) < 0) {
		error("Failed to format");
		return -1;
	}

	ret = access(path, F_OK);
	free(path);

	if (ret == 0) {
		return 1;

	} else if (errno == ENOENT) {
		return 0;

	} else {
		eerror("Failed to access");
		return -1;
	}
}

static int gpio_export(int id)
{
	int exported, fd;
	char *buf = NULL;
	ssize_t len;

	if ((exported = gpio_is_exported(id)) < 0) {
		error("Failed to check export status");
		return -1;

	} else if (exported == 1) {
		return 0;

	} else if ((fd = open(GPIO_SYS_EXPORT, O_WRONLY)) < 0) {
		eerror("Failed to open");
		return -1;

	} else if ((len = aprintf(&buf, "%u", id)) < 0) {
		error("Failed to format");
		close(fd);
		return -1;

	} else if (write(fd, buf, len) < 0) {
		eerror("Failed to write");
		free(buf);
		close(fd);
		return -1;
	}

	free(buf);
	close(fd);
	return 0;
}

static int gpio_unexport(int id)
{
	int exported, fd;
	char *buf = NULL;
	ssize_t len;

	if ((exported = gpio_is_exported(id)) < 0) {
		error("Failed to check export status");
		return -1;

	} else if (exported == 0) {
		return 0;

	} else if ((fd = open(GPIO_SYS_UNEXPORT, O_WRONLY)) < 0) {
		eerror("Failed to open");
		return -1;

	} else if ((len = aprintf(&buf, "%u", id)) < 0) {
		error("Failed to format");
		close(fd);
		return -1;

	} else if (write(fd, buf, len) < 0) {
		eerror("Failed to write");
		free(buf);
		close(fd);
		return -1;
	}

	free(buf);
	close(fd);
	return 0;
}

static int gpio_open_path(const char *format, int id)
{
	char *path = NULL;
	int fd;

	if (aprintf(&path, format, id) < 0) {
		error("Failed to format");
		return -1;

	} else if ((fd = open(path, O_RDWR)) < 0) {
		eerror("Failed to open");
		free(path);
		return -1;
	}

	free(path);
	return fd;
}

void *gpio_open(int id)
{
	struct gpio *gpio = NULL;

	if (id < 0) {
		error("Invalid id");
		return NULL;

	} else if ((gpio = malloc(sizeof(*gpio))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;
	}

	gpio->id = id;

	if (gpio_export(id) < 0) {
		error("Failed to export gpio");
		free(gpio);
		return NULL;

	} else if ((gpio->direction = gpio_open_path(GPIO_SYS_DIRECTION, id)) < 0) {
		error("Failed to export gpio");
		gpio_unexport(id);
		free(gpio);
		return NULL;

	} else if ((gpio->value = gpio_open_path(GPIO_SYS_VALUE, id)) < 0) {
		error("Failed to export gpio");
		close(gpio->direction);
		gpio_unexport(id);
		free(gpio);
		return NULL;
	}

	return gpio;
}

int gpio_close(void *device)
{
	struct gpio *gpio = device;
	int status = 0;

	if (gpio == NULL) {
		error("Invalid device");
		return -1;
	}

	if (gpio_set_mode(gpio, GPIO_INPUT) < 0) {
		warning("Failed to set mode");
		status = -1;
	}

	if (close(gpio->direction) < 0) {
		ewarning("Failed to close direction");
		status = -1;
	}

	if (close(gpio->value) < 0) {
		ewarning("Failed to close value");
		status = -1;
	}

	if (gpio_unexport(gpio->id) < 0) {
		warning("Failed to unexport");
		status = -1;
	}

	free(gpio);
	return status;
}

int gpio_get_mode(void *device)
{
	struct gpio *gpio = device;
	char mode[3];

	if (gpio == NULL) {
		error("Invalid device");
		return -1;

	} else if (read(gpio->direction, mode, sizeof(mode)) < 0) {
		eerror("Failed to read");
		return -1;

	} else if (strncmp(mode, "in", 2) == 0) {
		return GPIO_INPUT;

	} else if (strncmp(mode, "out", 3) == 0) {
		return GPIO_OUTPUT;
	}

	return -1;
}

int gpio_set_mode(void *device, int mode)
{
	struct gpio *gpio = device;

	if (gpio == NULL) {
		error("Invalid device");
		return -1;

	} else if (write(gpio->direction,
			 (mode & GPIO_HIGH) ? "high" : "low",
			 (mode & GPIO_HIGH) ? 4 : 3) < 0) {
		eerror("Failed to write");
		return -1;

	} else if (lseek(gpio->direction, 0, SEEK_SET) < 0) {
		eerror("Failed to reposition offset");
		return -1;

	} else if (write(gpio->direction,
			 (mode & GPIO_OUTPUT) ? "out" : "in",
			 (mode & GPIO_OUTPUT) ? 3 : 2) < 0) {
		eerror("Failed to write");
		return -1;
	}

	return 0;
}

int gpio_get_value(void *device)
{
	struct gpio *gpio = device;
	char value[1];

	if (gpio == NULL) {
		error("Invalid device");
		return -1;

	} else if (read(gpio->value, value, sizeof(value)) < 0) {
		eerror("Failed to read");
		return -1;

	} else if (*value == '1') {
		return GPIO_HIGH;

	} else if (*value == '0') {
		return GPIO_LOW;
	}

	return -1;
}

int gpio_set_value(void *device, int value)
{
	struct gpio *gpio = device;

	if (gpio == NULL) {
		error("Invalid device");
		return -1;

	} else if (write(gpio->value, (value & GPIO_HIGH) ? "1" : "0", 1) < 0) {
		eerror("Failed to write");
		return -1;
	}

	return 0;
}
