// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "sx1239.h"

#include <arpa/inet.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "utils/clock.h"
#include "utils/logger.h"
#include "utils/memory.h"
#include "utils/spi.h"

#include "gpio.h"

struct device {
	void *enable;
	void *reset;
	int spi;
};

#define SPI_SPEED 10000000
#define SPI_MODE SPI_MODE_0
#define SPI_BITS 8
#define SPI_LSB 0

#define REG_ADDR_MASK 0x7F
#define REG_READ 0x00
#define REG_WRITE 0x80

#define REG_OP_MODE 0x01
#define REG_DATA_MODUL 0x02
#define REG_BITRATE 0x03
#define REG_FRF 0x07
#define REG_LNA 0x18
#define REG_RX_BW 0x19
#define REG_RSSI_CONFIG 0x23
#define REG_RSSI_VALUE 0x24
#define REG_IRQ_FLAGS1 0x27
#define REG_TEST_LNA 0x58
#define REG_TEST_TCXO 0x59

#define REG_OP_MODE_MODE_MASK 0x1C
#define REG_OP_MODE_MODE_FS 0x08
#define REG_OP_MODE_MODE_RX 0x10

#define REG_DATA_MODUL_MODULATION_MASK 0x18
#define REG_DATA_MODUL_MODULATION_FSK 0x00
#define REG_DATA_MODUL_MODE_MASK 0x60
#define REG_DATA_MODUL_MODE_CONTINIOUS 0x60

#define REG_LNA_ZIN_MASK 0x80
#define REG_LNA_ZIN_200_OHMS 0x80
#define REG_LNA_GAIN_MASK 0x07
#define REG_LNA_GAIN_G1 0x01
#define REG_LNA_GAIN_G2 0x02
#define REG_LNA_GAIN_G3 0x03
#define REG_LNA_GAIN_G4 0x04
#define REG_LNA_GAIN_G5 0x05
#define REG_LNA_GAIN_G6 0x06

#define REG_RX_BW_DCC_FREQ_MASK 0xE0
#define REG_RX_BW_DCC_FREQ_16 0x00
#define REG_RX_BW_DCC_FREQ_8 0x20
#define REG_RX_BW_DCC_FREQ_4 0x40
#define REG_RX_BW_DCC_FREQ_2 0x60
#define REG_RX_BW_DCC_FREQ_1 0x80
#define REG_RX_BW_DCC_FREQ_1_2 0xA0
#define REG_RX_BW_DCC_FREQ_1_4 0xC0
#define REG_RX_BW_DCC_FREQ_1_8 0xE0
#define REG_RX_BW_RX_BW_MANT_MASK 0x18
#define REG_RX_BW_RX_BW_MANT_16 0x00
#define REG_RX_BW_RX_BW_MANT_20 0x08
#define REG_RX_BW_RX_BW_MANT_24 0x10
#define REG_RX_BW_RX_BW_EXP_MASK 0x07

#define REG_RSSI_CONFIG_START 0x01
#define REG_RSSI_CONFIG_DONE 0x02

#define REG_IRQ_FLAGS1_PLL_LOCK 0x10
#define REG_IRQ_FLAGS1_RX_READY 0x40

#define REG_TEST_LNA_HIGH_SENSITIVITY 0x2D

#define REG_TEST_TCXO_INPUT_MASK 0x10
#define REG_TEST_TCXO_INPUT_XTA 0x10

#define FSTEP 61.03515625
#define FXOSC 32000000

static int gpio_deinit(struct device *dev)
{
	int status = 0;

	if (dev->enable != NULL && gpio_close(dev->enable) < 0) {
		warning("Failed to close gpio enable");
		status = -1;
	}

	if (dev->reset != NULL && gpio_close(dev->reset) < 0) {
		warning("Failed to close gpio reset");
		status = -1;
	}

	return status;
}

static int gpio_init(struct device *dev, int enable, int reset)
{
	if ((dev->enable = gpio_open(enable)) == NULL) {
		error("Failed to open gpio enable");
		return -1;

	} else if (gpio_set_mode(dev->enable, GPIO_OUTPUT | GPIO_HIGH) < 0) {
		error("Failed to configure gpio enable");
		gpio_deinit(dev);
		return -1;

	} else if ((dev->reset = gpio_open(reset)) == NULL) {
		error("Failed to open gpio reset");
		gpio_deinit(dev);
		return -1;

	} else if (gpio_set_mode(dev->reset, GPIO_OUTPUT | GPIO_LOW) < 0) {
		error("Failed to configure gpio reset");
		gpio_deinit(dev);
		return -1;
	}

	sleep_ms(10);

	if (gpio_set_value(dev->reset, GPIO_HIGH) < 0) {
		error("Failed to set gpio reset");
		gpio_deinit(dev);
		return -1;
	}

	sleep_ms(10);

	if (gpio_set_value(dev->reset, GPIO_LOW) < 0) {
		error("Failed to reset gpio reset");
		gpio_deinit(dev);
		return -1;
	}

	return 0;
}

static int reg_msg(struct device *dev, const void *mosi, void *miso, size_t len)
{
	return spi_msg(dev->spi, mosi, miso, len);
}

static int reg_read(struct device *dev, uint8_t addr, uint8_t *buf, size_t len)
{
	uint8_t *mosi, *miso;

	if ((addr & ~REG_ADDR_MASK) == ~REG_ADDR_MASK) {
		error("Invalid address");
		return -1;

	} else if ((mosi = malloc(len + 1)) == NULL) {
		eerror("Failed to allocate memory");
		return -1;

	} else if ((miso = malloc(len + 1)) == NULL) {
		eerror("Failed to allocate memory");
		free(mosi);
		return -1;
	}

	memset(mosi, 0, len + 1);
	memset(miso, 0, len + 1);
	mosi[0] = REG_READ | (addr & REG_ADDR_MASK);

	if (reg_msg(dev, mosi, miso, len + 1) < 0) {
		error("Failed to send spi message");
		free(mosi);
		free(miso);
		return -1;
	}

	memcpy(buf, miso + 1, len);
	free(mosi);
	free(miso);
	return 0;
}

static int reg_write(struct device *dev, uint8_t addr, const uint8_t *buf,
		     size_t len)
{
	uint8_t *mosi, *miso;

	if ((addr & ~REG_ADDR_MASK) == ~REG_ADDR_MASK) {
		error("Invalid address");
		return -1;

	} else if ((mosi = malloc(len + 1)) == NULL) {
		eerror("Failed to allocate memory");
		return -1;

	} else if ((miso = malloc(len + 1)) == NULL) {
		eerror("Failed to allocate memory");
		free(mosi);
		return -1;
	}

	memset(mosi, 0, len + 1);
	memset(miso, 0, len + 1);
	mosi[0] = REG_WRITE | (addr & REG_ADDR_MASK);
	memcpy(mosi + 1, buf, len);

	if (reg_msg(dev, mosi, miso, len + 1) < 0) {
		error("Failed to send spi message");
		free(mosi);
		free(miso);
		return -1;
	}

	free(mosi);
	free(miso);
	return 0;
}

static int reg_get(struct device *dev, uint8_t addr)
{
	uint8_t val;

	if (reg_read(dev, addr, &val, 1) < 0) {
		error("Failed to read register");
		return -1;
	}

	return val;
}

static int reg_set_full(struct device *dev, uint8_t addr, uint8_t val)
{
	if (reg_write(dev, addr, &val, 1) < 0) {
		error("Failed to write register");
		return -1;
	}

	return 0;
}

static int reg_set_mask(struct device *dev, uint8_t addr, uint8_t val,
			uint8_t mask)
{
	int old, new;

	if ((old = reg_get(dev, addr)) < 0) {
		error("Failed to get register");
		return -1;
	}

	new = old;
	new &= ~mask;
	new |= val &mask;

	if (new != old && reg_set_full(dev, addr, new) < 0) {
		error("Failed to set full register");
		return -1;
	}

	return 0;
}

#define _sfx_reg_set(dev, addr, val, mask, sfx, ...) sfx
#define sfx_reg_set(args...) _sfx_reg_set(args, _mask, _full)
#define __reg_set(sfx, args...) reg_set##sfx(args)
#define _reg_set(sfx, args...) __reg_set(sfx, args)
#define reg_set(args...) _reg_set(sfx_reg_set(args), args)

static int reg_check(struct device *dev, uint8_t addr, uint8_t val,
		     uint8_t mask)
{
	int count, current;

	for (count = 0; count < 100; count++) {
		if ((current = reg_get(dev, addr)) < 0) {
			error("Failed to get register");
			return -1;

		} else if ((current & mask) == val) {
			return 0;

		} else {
			sleep_us(10);
		}
	}

	error("Failed to check register bit: Timeout");
	return -1;
}

static int set_op_mode(struct device *dev, uint8_t mode)
{
	if (reg_set(dev, REG_OP_MODE, mode, REG_OP_MODE_MODE_MASK) < 0) {
		error("Failed to set register");
		return -1;

	} else if (reg_check(dev, REG_OP_MODE, mode,
			     REG_OP_MODE_MODE_MASK) < 0) {
		error("Failed to check mode");
		return -1;
	}

	return 0;
}

static int set_freq(struct device *dev, uint32_t freq)
{
	union {
		uint32_t dword;
		uint8_t bytes[4];
	} f;

	f.dword = htonl(0.5 + freq / FSTEP);

	if (reg_write(dev, REG_FRF, &f.bytes[1], 3) < 0) {
		error("Failed to write RF carrier frequency");
		return -1;
	}

	return 0;
}

int sx1239_set_br(void *device, uint32_t br)
{
	struct device *dev = device;

	union {
		uint16_t word;
		uint8_t bytes[2];
	} bitrate;

	bitrate.word = htons((FXOSC + br / 2) / br);

	if (dev == NULL) {
		error("Invalid device");
		return -1;

	} else if (reg_write(dev, REG_BITRATE, bitrate.bytes, 2) < 0) {
		error("Failed to write bit rate");
		return -1;
	}

	return 0;
}

int sx1239_set_dcc_freq(void *device, uint8_t df)
{
	struct device *dev = device;
	uint8_t dcc_freq;

	switch (df) {
	case 8:
		dcc_freq = REG_RX_BW_DCC_FREQ_16;
		break;
	case 7:
		dcc_freq = REG_RX_BW_DCC_FREQ_8;
		break;
	case 6:
		dcc_freq = REG_RX_BW_DCC_FREQ_4;
		break;
	case 5:
		dcc_freq = REG_RX_BW_DCC_FREQ_2;
		break;
	case 4:
		dcc_freq = REG_RX_BW_DCC_FREQ_1;
		break;
	case 3:
		dcc_freq = REG_RX_BW_DCC_FREQ_1_2;
		break;
	case 2:
		dcc_freq = REG_RX_BW_DCC_FREQ_1_4;
		break;
	case 1:
	default:
		dcc_freq = REG_RX_BW_DCC_FREQ_1_8;
		break;
	}

	if (reg_set(dev, REG_RX_BW, dcc_freq,
		    REG_RX_BW_DCC_FREQ_MASK) < 0) {
		error("Failed to configure DCC cutoff frequency");
		return -1;
	}

	return 0;
}

int sx1239_set_lna_gain(void *device, uint8_t lg)
{
	struct device *dev = device;
	uint8_t lna_gain;

	switch (lg) {
	case 6:
		lna_gain = REG_LNA_GAIN_G6;
		break;
	case 5:
		lna_gain = REG_LNA_GAIN_G5;
		break;
	case 4:
		lna_gain = REG_LNA_GAIN_G4;
		break;
	case 3:
		lna_gain = REG_LNA_GAIN_G3;
		break;
	case 2:
		lna_gain = REG_LNA_GAIN_G2;
		break;
	case 1:
	default:
		lna_gain = REG_LNA_GAIN_G1;
		break;
	}

	if (reg_set(dev, REG_LNA, lna_gain,
		    REG_LNA_GAIN_MASK) < 0) {
		error("Failed to configure LNA gain");
		return -1;
	}

	return 0;
}

static int dev_init(struct device *dev)
{
	if (sx1239_set_dcc_freq(dev, 0) < 0) {
		error("Failed to configure default DCC cutoff frequency");
		return -1;

	} else if (sx1239_set_lna_gain(dev, 0) < 0) {
		error("Failed to configure default LNA gain");
		return -1;

	} else if (reg_set(dev, REG_TEST_LNA,
			   REG_TEST_LNA_HIGH_SENSITIVITY) < 0) {
		error("Failed to configure sensitivity");
		return -1;

	} else if (reg_set(dev, REG_TEST_TCXO,
			   REG_TEST_TCXO_INPUT_XTA,
			   REG_TEST_TCXO_INPUT_MASK) < 0) {
		error("Failed to configure oscillator");
		return -1;

	} else if (reg_set(dev, REG_LNA,
			   REG_LNA_ZIN_200_OHMS,
			   REG_LNA_ZIN_MASK) < 0) {
		error("Failed to configure LNA");
		return -1;

	} else if (reg_set(dev, REG_DATA_MODUL,
			   REG_DATA_MODUL_MODE_CONTINIOUS |
			       REG_DATA_MODUL_MODULATION_FSK,
			   REG_DATA_MODUL_MODE_MASK |
			       REG_DATA_MODUL_MODULATION_MASK) < 0) {
		error("Failed to configure modulation");
		return -1;
	}

	return 0;
}

static struct {
	uint8_t mant_exp;
	uint32_t bw_fsk;
	uint32_t bw_ook;
} rxbw_regs[] = {
    {REG_RX_BW_RX_BW_MANT_24 | 7, 2600, 1300},
    {REG_RX_BW_RX_BW_MANT_20 | 7, 3100, 1600},
    {REG_RX_BW_RX_BW_MANT_16 | 7, 3900, 2000},
    {REG_RX_BW_RX_BW_MANT_24 | 6, 5200, 2600},
    {REG_RX_BW_RX_BW_MANT_20 | 6, 6300, 3100},
    {REG_RX_BW_RX_BW_MANT_16 | 6, 7800, 3900},
    {REG_RX_BW_RX_BW_MANT_24 | 5, 10400, 5200},
    {REG_RX_BW_RX_BW_MANT_20 | 5, 12500, 6300},
    {REG_RX_BW_RX_BW_MANT_16 | 5, 15600, 7800},
    {REG_RX_BW_RX_BW_MANT_24 | 4, 20800, 10400},
    {REG_RX_BW_RX_BW_MANT_20 | 4, 25000, 12500},
    {REG_RX_BW_RX_BW_MANT_16 | 4, 31300, 15600},
    {REG_RX_BW_RX_BW_MANT_24 | 3, 41700, 20800},
    {REG_RX_BW_RX_BW_MANT_20 | 3, 50000, 25000},
    {REG_RX_BW_RX_BW_MANT_16 | 3, 62500, 31300},
    {REG_RX_BW_RX_BW_MANT_24 | 2, 83300, 41700},
    {REG_RX_BW_RX_BW_MANT_20 | 2, 100000, 50000},
    {REG_RX_BW_RX_BW_MANT_16 | 2, 125000, 62500},
    {REG_RX_BW_RX_BW_MANT_24 | 1, 166700, 83300},
    {REG_RX_BW_RX_BW_MANT_20 | 1, 200000, 100000},
    {REG_RX_BW_RX_BW_MANT_16 | 1, 250000, 125000},
    {REG_RX_BW_RX_BW_MANT_24 | 0, 333300, 166700},
    {REG_RX_BW_RX_BW_MANT_20 | 0, 400000, 200000},
    {REG_RX_BW_RX_BW_MANT_16 | 0, 500000, 250000},
};

int sx1239_set_bw(void *device, uint32_t bw)
{
	struct device *dev = device;
	uint8_t mant_exp;
	size_t i;
	int mode;

	if (dev == NULL) {
		error("Invalid device");
		return -1;

	} else if ((mode = reg_get(dev, REG_DATA_MODUL)) < 0) {
		error("Failed to get mode");
		return -1;
	}

	for (mant_exp = 0, i = 0; i < sizeofa(rxbw_regs); i++) {
		if ((mode & REG_DATA_MODUL_MODULATION_MASK) ==
		    REG_DATA_MODUL_MODULATION_FSK) {
			if (bw <= 2 * rxbw_regs[i].bw_fsk) {
				mant_exp = rxbw_regs[i].mant_exp;
				break;
			}

		} else {
			if (bw <= 2 * rxbw_regs[i].bw_ook) {
				mant_exp = rxbw_regs[i].mant_exp;
				break;
			}
		}
	}

	if (mant_exp == 0) {
		error("Invalid bandwidth");
		return -1;

	} else if (reg_set(dev, REG_RX_BW, mant_exp,
			   REG_RX_BW_RX_BW_MANT_MASK |
			       REG_RX_BW_RX_BW_EXP_MASK) < 0) {
		error("Failed to set bandwidth");
		return -1;
	}

	return 0;
}

int sx1239_start_rx(void *device, uint32_t freq)
{
	struct device *dev = device;

	if (dev == NULL) {
		error("Invalid device");
		return -1;

	} else if (set_freq(dev, freq) < 0) {
		error("Failed to set RF carrier frequency");
		return -1;

	} else if (set_op_mode(dev, REG_OP_MODE_MODE_FS) < 0) {
		error("Failed to set FS mode");
		return -1;

	} else if (set_op_mode(dev, REG_OP_MODE_MODE_RX) < 0) {
		error("Failed to set RX mode");
		return -1;

	} else if (reg_check(dev, REG_IRQ_FLAGS1,
			     REG_IRQ_FLAGS1_RX_READY,
			     REG_IRQ_FLAGS1_RX_READY) < 0) {
		error("Failed to wait for RX ready");
		return -1;
	}

	return 0;
}

int sx1239_rssi(void *device)
{
	struct device *dev = device;
	int rssi;

	if (dev == NULL) {
		error("Invalid device");
		return -1;

	} else if (reg_set(dev, REG_RSSI_CONFIG, REG_RSSI_CONFIG_START) < 0) {
		error("Failed to start RSSI measurement");
		return -1;

	} else if (reg_check(dev, REG_RSSI_CONFIG,
			     REG_RSSI_CONFIG_DONE,
			     REG_RSSI_CONFIG_DONE) < 0) {
		error("Failed to wait for RSSI done");
		return -1;

	} else if ((rssi = reg_get(dev, REG_RSSI_VALUE)) < 0) {
		error("Failed to get RSSI");
		return -1;
	}

	return rssi;
}

void *sx1239_open(const char *spidev, int enable, int reset)
{
	struct device *dev;

	if ((dev = malloc(sizeof(*dev))) == NULL) {
		eerror("Failed to allocate memory");
		return NULL;

	} else if (gpio_init(dev, enable, reset) < 0) {
		error("Failed to initialize gpio");
		free(dev);
		return NULL;

	} else if ((dev->spi = spi_init(spidev, SPI_SPEED, SPI_MODE,
					SPI_BITS, SPI_LSB)) < 0) {
		error("Failed to initialize spi");
		gpio_deinit(dev);
		free(dev);
		return NULL;

	} else if (dev_init(dev) < 0) {
		error("Failed to initialize device");
		gpio_deinit(dev);
		close(dev->spi);
		free(dev);
		return NULL;
	}

	return dev;
}

int sx1239_close(void *device)
{
	struct device *dev = device;
	int status = 0;

	if (dev == NULL) {
		error("Invalid device");
		return -1;
	}

	if (gpio_deinit(dev) < 0) {
		warning("Failed to deinitialize gpio");
		status = -1;
	}

	if (close(dev->spi) < 0) {
		warning("Failed to close spi");
		status = -1;
	}

	free(dev);
	return status;
}
