// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "core_fscan.h"

#include <stdlib.h>
#include <sys/types.h>

#if defined(PLATFORM_WSTA)
#include "utils/clock.h"
#endif
#include "utils/logger.h"
#include "utils/memory.h"

#include "config.h"
#include "core.h"

#if defined(PLATFORM_WIFC)
#include "drivers/sx1239.h"
#endif

#include "lib/loragw/lbt.h"
#if defined(PLATFORM_WSTA)
#include "lib/loragw/aux.h"
#include "lib/loragw/fpga.h"
#include "lib/loragw/hal.h"
#include "lib/loragw/radio.h"
#include "lib/loragw/reg.h"
#endif

#if defined(PLATFORM_WIFC)
#define SNIFFER_NB_MAX 1
#define SNIFFER_SPIDEV "/dev/spidev1.0"
#define SNIFFER_GPIOEN 10
#define SNIFFER_GPIORST 135
#define SNIFFER_BANDWIDTH 200000
#define SNIFFER_BITRATE 600
#define SNIFFER_LNA_GAIN 1
#define SNIFFER_DCC_FREQ 1
#endif

#if defined(PLATFORM_WSTA)
#define SNIFFER_NB_MAX 1
#define SNIFFER_BANDWIDTH LGW_SX127X_RXBW_100K_HZ

#define RSSI_MEASURE_NB 100
#define RSSI_RANGE 256

#define FREQ_TO_FREQ_REG(freq) (int32_t)(((uint64_t)freq << 19) / (uint64_t)32000000)
#endif

#if defined(PLATFORM_WIFC)
struct sniffer_desc {
	const char *spi_name;
	void *handle;
};
#endif

#if defined(PLATFORM_WIFC)
static struct sniffer_desc sniffer_list[SNIFFER_NB_MAX] = {
    {.spi_name = SNIFFER_SPIDEV,
     .handle = NULL}};
#endif

struct fscan_core *fscan_core_init(void)
{
	struct fscan_core *fscan_core = NULL;

	if ((fscan_core = zalloc(sizeof(*fscan_core))) == NULL) {
		eerror("Failed to allocate fscan core");
		return NULL;
	}

	return fscan_core;
}

int fscan_core_start(uint32_t receiver_id)
{
	if (lbt_is_enabled()) {
		notice("LBT is enabled, no frequency scan available");
		return -1;
	}
#if defined(PLATFORM_WIFC)
	/* Initialize sniffer */
	else if ((sniffer_list[receiver_id].handle = sx1239_open(sniffer_list[receiver_id].spi_name,
								 SNIFFER_GPIOEN, SNIFFER_GPIORST)) == NULL) {
		error("Failed to open receiver id %u", receiver_id);
		return -1;

	} else if (sx1239_set_bw(sniffer_list[receiver_id].handle, SNIFFER_BANDWIDTH) < 0) {
		error("Failed to configure bandwidth");
		fscan_core_shutdown(receiver_id);
		return -1;

	} else if (sx1239_set_br(sniffer_list[receiver_id].handle, SNIFFER_BITRATE) < 0) {
		error("Failed to configure bit rate");
		fscan_core_shutdown(receiver_id);
		return -1;
	} else if (sx1239_set_dcc_freq(sniffer_list[receiver_id].handle, SNIFFER_DCC_FREQ) < 0) {
		error("Failed to configure DCC frequency");
		fscan_core_shutdown(receiver_id);
		return -1;
	} else if (sx1239_set_lna_gain(sniffer_list[receiver_id].handle, SNIFFER_LNA_GAIN) < 0) {
		error("Failed to configure LNA gain");
		fscan_core_shutdown(receiver_id);
		return -1;
	}
#endif
#if defined(PLATFORM_WSTA)
	/* Unused parameter */
	receiver_id = receiver_id;
#endif

	return 0;
}

void fscan_core_shutdown(uint32_t receiver_id)
{
#if defined(PLATFORM_WIFC)
	if (sniffer_list[receiver_id].handle) {
		sx1239_close(sniffer_list[receiver_id].handle);
	}
#endif
#if defined(PLATFORM_WSTA)
	/* Unused parameter */
	receiver_id = receiver_id;

/* Nothing to shutdown for this platform */
#endif
}

void fscan_core_free(struct core *core)
{
	if (core && core->fscan && core->fscan->core) {
		free(core->fscan->core);
	}
}

int fscan_get_capabilities(struct core *core, struct sniffer_caps *caps)
{
	(void)core;

	caps->sniffer_nb = SNIFFER_NB_MAX;

	return 0;
}

int fscan_do_scan(float *rssi, struct core *core, uint32_t receiver_id, uint32_t frequency)
{
#if defined(PLATFORM_WIFC)
	int sx1239_rssi_value;
#endif
#if defined(PLATFORM_WSTA)
	int r;
	int32_t reg_val;
	uint8_t read_burst[RSSI_RANGE * 2];
	uint16_t rssi_histo;
	uint16_t rssi_nb_total = 0;
	uint32_t rssi_sum = 0;
	unsigned int nb_loop = 0;
#endif

	if (receiver_id >= SNIFFER_NB_MAX) {
		error("Invalid receiver id %u", receiver_id);
		return -1;
	}

#if defined(PLATFORM_WIFC)
	/* Update frequency scan */
	if (sx1239_start_rx(sniffer_list[receiver_id].handle, frequency) < 0) {
		error("Failed to update frequency");
		return -1;
	}

	/* Retrieve scan result */
	if ((sx1239_rssi_value = sx1239_rssi(sniffer_list[receiver_id].handle)) < 0) {
		error("Failed to get rssi");
		return -1;
	}

	*rssi = -((float)sx1239_rssi_value / 2) + core->config->core->lbt.rssi_offset;
#endif

#if defined(PLATFORM_WSTA)
	/* Set number of measure */
	if (lgw_fpga_reg_w(LGW_FPGA_HISTO_NB_READ, RSSI_MEASURE_NB - 1) != LGW_REG_SUCCESS) {
		error("Failed to configure number of RSSI measure in FPGA");
		return -1;
	}

	/* Initialize frequency */
	lgw_fpga_reg_w(LGW_FPGA_HISTO_SCAN_FREQ, FREQ_TO_FREQ_REG(frequency));

	/* Set SX127x */
	if (lgw_setup_sx127x(frequency, MOD_FSK, SNIFFER_BANDWIDTH, core->config->core->lbt.rssi_offset) != 0) {
		error("SX127x setup failed");
		return -1;
	}

	/* Start FPGA state machine for spectral scan */
	lgw_fpga_reg_w(LGW_FPGA_CTRL_FEATURE_START, 1);

	/* Clean histogram */
	lgw_fpga_reg_w(LGW_FPGA_CTRL_CLEAR_HISTO_MEM, 1);

	/* Wait for histogram clean to start */
	do {
		lgw_fpga_reg_r(LGW_FPGA_STATUS, &reg_val);
		sleep_ms(1);
		nb_loop++;
	} while (((TAKE_N_BITS_FROM((uint8_t)reg_val, 0, 5)) != 1) && (nb_loop < 100)); /* Clear has started */

	/* Set scan frequency during clear process */
	/* We can directly set the scan frequency */
	lgw_fpga_reg_w(LGW_FPGA_HISTO_SCAN_FREQ, FREQ_TO_FREQ_REG(frequency));

	/* Release FPGA state machine */
	lgw_fpga_reg_w(LGW_FPGA_CTRL_CLEAR_HISTO_MEM, 0);

	/* Wait for histogram ready */
	nb_loop = 0;
	do {
		lgw_fpga_reg_r(LGW_FPGA_STATUS, &reg_val);
		sleep_ms(1);
		nb_loop++;
	} while (((TAKE_N_BITS_FROM((uint8_t)reg_val, 5, 1)) != 1) && (nb_loop < 100));

	/* Stop FPGA state machine for spectral scan */
	lgw_fpga_reg_w(LGW_FPGA_CTRL_FEATURE_START, 0);

	/* Read histogram */
	lgw_fpga_reg_w(LGW_FPGA_CTRL_ACCESS_HISTO_MEM, 1); /* Host gets access to FPGA RAM */
	lgw_fpga_reg_w(LGW_FPGA_HISTO_RAM_ADDR, 0);
	lgw_fpga_reg_rb(LGW_FPGA_HISTO_RAM_DATA, read_burst, RSSI_RANGE * 2);
	lgw_fpga_reg_w(LGW_FPGA_CTRL_ACCESS_HISTO_MEM, 0); /* FPGA gets access to RAM back */

	for (r = 0; r < RSSI_RANGE; r++) {
		rssi_histo = (uint16_t)read_burst[2 * r] | ((uint16_t)read_burst[(2 * r) + 1] << 8);
		rssi_nb_total += rssi_histo;
		rssi_sum += rssi_histo * r;
	}
	*rssi = -((float)rssi_sum) / (2 * rssi_nb_total);
#endif

	return 0;
}

bool fscan_is_lbt_enabled(uint32_t receiver_id)
{
	/* Unused parameter */
	receiver_id = receiver_id;

	if (lbt_is_enabled()) {
		return true;
	} else {
		return false;
	}
}
