// Copyright (c) 2018 Kerlink SA. All rights reserved.

#include "config.h"

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include "lib/sx1301ar/err.h"
#include "lib/sx1301ar/spi.h"

#include "utils/logger.h"
#include "utils/memory.h"

#define lerror(fmt, args...) error(fmt ": %s", ##args, sx1301ar_err_message(sx1301ar_errno))

#define CONFIG_SX1301AR_JSONOBJ "SX1301_array_conf"

static pthread_mutex_t spi_mutex[SX1301AR_MAX_BOARD_NB];
static int spi_fd[SX1301AR_MAX_BOARD_NB];

#define _declare_spi(func, nb, const...)                                                             \
	static int spi##nb##_##func(uint8_t h, uint16_t a, const uint8_t *d, uint32_t l, uint8_t *s) \
	{                                                                                            \
		int x;                                                                               \
		pthread_mutex_lock(&spi_mutex[nb - 1]);                                              \
		x = spi_linuxdev_##func(h, spi_fd[nb - 1], a, d, l, s);                              \
		pthread_mutex_unlock(&spi_mutex[nb - 1]);                                            \
		return x;                                                                            \
	}

#define declare_spi(nb) _declare_spi(read, nb) _declare_spi(write, nb, const)

declare_spi(1);
#if (SX1301AR_MAX_BOARD_NB > 1)
declare_spi(2);
#endif
#if (SX1301AR_MAX_BOARD_NB > 2)
declare_spi(3);
#endif
#if (SX1301AR_MAX_BOARD_NB > 3)
declare_spi(4);
#endif

#define JSON_PARSE_BOOLEAN(object, name, value)                                                            \
	{                                                                                                  \
		JSON_Value *_value = NULL;                                                                 \
		if ((_value = json_object_dotget_value(object, name)) != NULL) {                           \
			if (json_value_get_type(_value) == JSONBoolean) {                                  \
				value = json_value_get_boolean(_value);                                    \
				debug("Value \"" name "\" configured to: %s", (value) ? "true" : "false"); \
			} else {                                                                           \
				warning("Wrong data type for " name);                                      \
			}                                                                                  \
		}                                                                                          \
	}

#define JSON_PARSE_NUMBER(object, name, value)                                                \
	{                                                                                     \
		JSON_Value *_value = NULL;                                                    \
		if ((_value = json_object_dotget_value(object, name)) != NULL) {              \
			if (json_value_get_type(_value) == JSONNumber) {                      \
				value = json_value_get_number(_value);                        \
				debug("Value \"" name "\" configured to: %g", (double)value); \
			} else {                                                              \
				warning("Wrong data type for " name);                         \
			}                                                                     \
		}                                                                             \
	}

static int parse_rfchain_configuration(struct core *core, JSON_Array *rfchain_array, uint8_t nb_board)
{
	struct config_core *config_core = core->config->core;
	sx1301ar_board_cfg_t *board = &config_core->boards[nb_board].board;
	struct config_rfchain *rfchains = config_core->boards[nb_board].rfchains;

	size_t nb_rfchain, nb_lut;
	JSON_Object *txgain_obj;
	JSON_Object *rfchain_obj;
	JSON_Array *txlut_array;
	JSON_Value *val;

	for (nb_rfchain = 0;
	     nb_rfchain < json_array_get_count(rfchain_array) &&
	     nb_rfchain < SX1301AR_BOARD_RFCHAIN_NB;
	     nb_rfchain++) {

		/* Get RF chain conf object */
		rfchain_obj = json_array_get_object(rfchain_array, nb_rfchain);

		/* Parse configuration */
		board->rf_chain[nb_rfchain].rx_enable = (bool)json_object_get_boolean(rfchain_obj, "rx_enable");
		board->rf_chain[nb_rfchain].tx_enable = (bool)json_object_get_boolean(rfchain_obj, "tx_enable");
		JSON_PARSE_NUMBER(rfchain_obj, "antenna_gain", board->rf_chain[nb_rfchain].antenna_gain);
		JSON_PARSE_NUMBER(rfchain_obj, "insertion_loss", board->rf_chain[nb_rfchain].insertion_loss);
		board->rf_chain[nb_rfchain].rssi_offset = (float)json_object_get_number(rfchain_obj, "rssi_offset");

		val = json_object_get_value(rfchain_obj, "rssi_offset_coeff_a");
		if (json_value_get_type(val) == JSONNumber)
			board->rf_chain[nb_rfchain].rssi_offset_coeff_a = (int16_t)json_value_get_number(val);

		val = json_object_get_value(rfchain_obj, "rssi_offset_coeff_b");
		if (json_value_get_type(val) == JSONNumber)
			board->rf_chain[nb_rfchain].rssi_offset_coeff_b = (int16_t)json_value_get_number(val);

		info("rf_chain %zd: RSSI offset configured at %.1f, a=%d, b=%d", nb_rfchain,
		     board->rf_chain[nb_rfchain].rssi_offset,
		     board->rf_chain[nb_rfchain].rssi_offset_coeff_a,
		     board->rf_chain[nb_rfchain].rssi_offset_coeff_b);

		/* Record TX limits if given */
		JSON_PARSE_NUMBER(rfchain_obj, "tx_freq_min", rfchains[nb_rfchain].tx_freq_min);
		JSON_PARSE_NUMBER(rfchain_obj, "tx_freq_max", rfchains[nb_rfchain].tx_freq_max);

		/* set configuration for tx gains */
		txlut_array = json_object_get_array(rfchain_obj, "tx_lut");
		if (txlut_array != NULL) {
			board->rf_chain[nb_rfchain].tx_lut.size = json_array_get_count(txlut_array);
			info("rf_chain %zd: Configuring Tx LUT with %u indexes", nb_rfchain, board->rf_chain[nb_rfchain].tx_lut.size);

			for (nb_lut = 0;
			     nb_lut < board->rf_chain[nb_rfchain].tx_lut.size &&
			     nb_lut < SX1301AR_BOARD_MAX_LUT_NB;
			     nb_lut++) {

				/* Get TX gain object from LUT */
				txgain_obj = json_array_get_object(txlut_array, nb_lut);

				/* rf power */
				val = json_object_dotget_value(txgain_obj, "rf_power");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].rf_power = (int8_t)json_value_get_number(val);
				} else {
					warning("Data type for %s[%zd] seems wrong, please check", "rf_power", nb_lut);
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].rf_power = 0;
				}

				/* FPGA dig gain */
				val = json_object_dotget_value(txgain_obj, "fpga_dig_gain");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].fpga_dig_gain = (uint8_t)json_value_get_number(val);
				} else {
					warning("Data type for %s[%zd] seems wrong, please check", "fpga_dig_gain", nb_lut);
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].fpga_dig_gain = 0;
				}

				/* AD9361 atten */
				val = json_object_get_value(txgain_obj, "ad9361_atten");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_gain.atten = (uint16_t)json_value_get_number(val);
				} else {
					warning("Data type for %s[%zd] seems wrong, please check", "ad9361_atten", nb_lut);
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_gain.atten = 0;
				}

				/* AD9361 auxdac vref lsb */
				val = json_object_dotget_value(txgain_obj, "ad9361_auxdac_vref");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_gain.auxdac_vref = (uint8_t)json_value_get_number(val);
				} else {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_gain.auxdac_vref = 0;
				}

				/* AD9361 auxdac word */
				val = json_object_dotget_value(txgain_obj, "ad9361_auxdac_word");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_gain.auxdac_word = (uint16_t)json_value_get_number(val);
				} else {
					warning("Data type for %s[%zd] seems wrong, please check", "ad9361_auxdac_word", nb_lut);
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_gain.auxdac_word = 0;
				}

				/* AD9361 temperature compensation coeff A */
				val = json_object_dotget_value(txgain_obj, "ad9361_tcomp_coeff_a");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_tcomp.coeff_a = (int16_t)json_value_get_number(val);
				} else {
					warning("Data type for %s[%zd] seems wrong, please check", "ad9361_tcomp_coeff_a", nb_lut);
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_tcomp.coeff_a = 0;
				}

				/* AD9361 temperature compensation coeff B */
				val = json_object_dotget_value(txgain_obj, "ad9361_tcomp_coeff_b");
				if (json_value_get_type(val) == JSONNumber) {
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_tcomp.coeff_b = (int16_t)json_value_get_number(val);
				} else {
					warning("Data type for %s[%zd] seems wrong, please check", "ad9361_tcomp_coeff_b", nb_lut);
					board->rf_chain[nb_rfchain].tx_lut.lut[nb_lut].ad9361_tcomp.coeff_b = 0;
				}
			}
		}
	}

	return 0;
}

static int parse_sx1301_single(struct core *core, JSON_Object *root, uint8_t nb_board)
{
	struct config_core *config_core = core->config->core;
	sx1301ar_board_cfg_t *board = &config_core->boards[nb_board].board;
	sx1301ar_lbt_cfg_t *lbt = &config_core->boards[nb_board].lbt;

	int i, j;
	size_t nb_chip, nb_chan;
	int x;		     /* return code for SX1301ar driver functions */
	char param_name[64]; /* used to generate variable parameter names */
	JSON_Object *chip_obj;
	JSON_Object *chan_obj;
	JSON_Object *lbt_obj;
	JSON_Object *lbtchan_obj;
	JSON_Array *sx1301_array;
	JSON_Array *rfchain_array;
	JSON_Array *lbtchan_array;
	JSON_Value *val;
	uint8_t idx_chan;
	uint32_t tmp_u32;
	int sf_a, sf_b;
	unsigned long long ull = 0, ull2 = 0;
	const char *str;
	int str_len;

	/* SPI interface */
	const char *spi_path = NULL;

	/* Mandatory: Type of the board */
	val = json_object_get_value(root, "board_type");
	if (json_value_get_type(val) == JSONString) {
		str = json_value_get_string(val);
		if (strncmp(str, "MASTER", strlen(str)) == 0) {
			info("Master board");
			board->board_type = BRD_MASTER;
		} else if (strncmp(str, "SLAVE", strlen(str)) == 0) {
			info("Slave board");
			board->board_type = BRD_SLAVE;
		} else {
			error("Board type %s is not valid", str);
			return -1;
		}
	} else if (board->board_type == BRD_TYPE_UNKNOWN) {
		error("Impossible to parse mandatory parameter 'board_type'");
		return -1;
	}

	/* Mandatory: Rx frequency of the board (ie. RX freq of the AD9361) */
	val = json_object_get_value(root, "board_rx_freq");
	if (json_value_get_type(val) == JSONNumber) {
		board->rx_freq_hz = (uint32_t)json_value_get_number(val);
		info("Board center frequency configured at %u Hz", board->rx_freq_hz);
	} else if (board->rx_freq_hz == 0) {
		error("Impossible to parse mandatory parameter 'board_rx_freq'");
		return -1;
	}

	/* Mandatory: Rx bandwidth of the board (ie. RX bw of the AD9361) */
	val = json_object_get_value(root, "board_rx_bw");
	if (json_value_get_type(val) == JSONNumber) {
		board->rx_bw_hz = (uint32_t)json_value_get_number(val);
		info("Board RX bandwidth configured at %u Hz", board->rx_bw_hz);
	} else if (board->rx_bw_hz == 0) {
		error("Impossible to parse mandatory parameter 'board_rx_bw'");
		return -1;
	}
	/* Optional: radio mode */
	JSON_PARSE_BOOLEAN(root, "full_duplex", board->full_duplex);

	/* RF chains configuration */
	rfchain_array = json_object_get_array(root, "rf_chain_conf");
	if (rfchain_array != NULL) {
		x = parse_rfchain_configuration(core, rfchain_array, nb_board);
	} else {
		warning("No rf_chain configured");
	}

	/* Listen-Before-Talk configuration */
	lbt_obj = json_object_get_object(root, "lbt_conf");
	if (lbt_obj == NULL) {
		info("No configuration for LBT");
	} else {
		val = json_object_get_value(lbt_obj, "enable");
		if (json_value_get_type(val) == JSONBoolean) {
			lbt->enable = (bool)json_value_get_boolean(val);
		} else {
			warning("Data type for lbt_conf.enable seems wrong, please check");
			lbt->enable = false;
		}
		if (lbt->enable == true) {
			val = json_object_get_value(lbt_obj, "rssi_target");
			if (json_value_get_type(val) == JSONNumber) {
				lbt->rssi_target = (int8_t)json_value_get_number(val);
			} else {
				warning("Data type for lbt_conf.rssi_target seems wrong, please check");
				lbt->rssi_target = 0;
			}
			val = json_object_get_value(lbt_obj, "rssi_shift");
			if (json_value_get_type(val) == JSONNumber) {
				lbt->rssi_shift = (int8_t)json_value_get_number(val);
			} else {
				warning("Data type for lbt_conf.rssi_shift seems wrong, please check");
				lbt->rssi_shift = 0;
			}
			/* set LBT channels configuration */
			lbtchan_array = json_object_get_array(lbt_obj, "chan_cfg");
			if (lbtchan_array != NULL) {
				lbt->nb_channel = json_array_get_count(lbtchan_array);
				info("%u LBT channels configured", lbt->nb_channel);
			}
			for (nb_chan = 0;
			     nb_chan < lbt->nb_channel &&
			     nb_chan < SX1301AR_LBT_CHANNEL_NB_MAX;
			     nb_chan++) {

				/* Get LBT channel configuration object from array */
				lbtchan_obj = json_array_get_object(lbtchan_array, nb_chan);

				/* Channel frequency */
				val = json_object_dotget_value(lbtchan_obj, "freq_hz");
				if (json_value_get_type(val) == JSONNumber) {
					lbt->channels[nb_chan].freq_hz = (uint32_t)json_value_get_number(val);
				} else {
					warning("Data type for lbt_cfg.channels[%zd].freq_hz seems wrong, please check", nb_chan);
					lbt->channels[nb_chan].freq_hz = 0;
				}

				/* Channel scan time */
				val = json_object_dotget_value(lbtchan_obj, "scan_time_us");
				if (json_value_get_type(val) == JSONNumber) {
					lbt->channels[nb_chan].scan_time_us = (uint16_t)json_value_get_number(val);
				} else {
					warning("Data type for lbt_cfg.channels[%zd].scan_time_us seems wrong, please check", nb_chan);
					lbt->channels[nb_chan].scan_time_us = 0;
				}

				/* Channel transmit duration when allowed */
				val = json_object_dotget_value(lbtchan_obj, "transmit_time_ms");
				if (json_value_get_type(val) == JSONNumber) {
					lbt->channels[nb_chan].transmit_time_ms = (uint16_t)json_value_get_number(val);
				} else {
					warning("Data type for lbt_cfg.channels[%zd].transmit_time_ms seems wrong, please check", nb_chan);
					lbt->channels[nb_chan].transmit_time_ms = 0;
				}
			}
		} else {
			info("LBT is disabled");
		}
	}

	/* Optional: FSK sync word */
	val = json_object_get_value(root, "FSK_sync");
	if (json_value_get_type(val) == JSONString) {
		str = json_value_get_string(val);
		str_len = strlen(str);
		if ((str_len < 2) || (str_len > 16) || (str_len % 2 != 0)) {
			error("Impossible to parse parameter 'FSK_sync'");
			return -1;
		}
		i = sscanf(str, "%llx", &ull);
		if (i != 1) {
			error("Failed to parse hex string for 'FSK_sync'");
			return -1;
		}
		board->fsk_sync_size = str_len / 2;
		board->fsk_sync_word = (uint64_t)ull;
	}

	/* Optional: enable cohabitation setting for LoRa MAC public networks */
	board->loramac_public = (bool)json_object_get_boolean(root, "loramac_public");

	/* Fine timestamp decryption enabled */
	JSON_PARSE_BOOLEAN(root, "fine_timestamp_decrypt_enable", board->fine_timestamp_decrypt_enable);

	/* Getting AES-128 decryption key */
	val = json_object_get_value(root, "aes_key");
	if (json_value_get_type(val) == JSONString) {
		str = json_value_get_string(val);
		str_len = strlen(str);
		if (str_len != 32) /* 16 bytes => 32 characters */
		{
			error("Wrong size for AES-128 key");
		} else {
			i = sscanf(json_value_get_string(val), "%16llx%16llx", &ull, &ull2);
			if (i != 2) {
				error("Failed to parse hex string for AES-128 key");
			} else {
				for (i = 0; i < 8; i++) {
					board->aes_key[i] = (uint8_t)((ull >> (56 - i * 8)) & 0xFF);
					board->aes_key[i + 8] = (uint8_t)((ull2 >> (56 - i * 8)) & 0xFF);
				}
				info("Gateway AES-128 key is configured to %016llX%016llX", ull, ull2);
			}
		}
	} else {
		error("Wrong format for AES-128 key");
	}

	/* Get room temperature used as Tref reference */
	board->room_temp_ref = (int8_t)json_object_get_number(root, "calibration_temperature_celsius_room");
	info("Calibration room temperature (Tref) set to %u oC", board->room_temp_ref);

	/* Get radio temperature sensor reference value (for Tref°C) */
	board->ad9361_temp_ref = (uint8_t)json_object_get_number(root, "calibration_temperature_code_ad9361");
	info("Calibration radio temperature code at [Tref] set to %u", board->ad9361_temp_ref);

	/* Get DSP status report interval */
	board->dsp_stat_interval = (uint8_t)json_object_get_number(root, "dsp_stat_interval");

	/* Get the number of dsp on the board */
	val = json_object_get_value(root, "nb_dsp");
	if (json_value_get_type(val) == JSONNumber) {
		board->nb_dsp = (uint32_t)json_value_get_number(val);
		info("%u dsp on the board", board->nb_dsp);
	} else if (board->nb_dsp == 0) {
		error("Impossible to parse mandatory parameter 'nb_dsp'");
		return -1;
	}

	/* Optional: Configure board to match fine timestamp even if packet has CRC error */
	val = json_object_get_value(root, "fts_match_crc_error");
	if (json_value_get_type(val) == JSONBoolean) {
		board->match_tmst_crc_err = (bool)json_value_get_boolean(val);
	}

	/* Optional: Get the encrypted/main fine timestamp version to be sent by DSP */
	val = json_object_get_value(root, "fts_version");
	if (json_value_get_type(val) == JSONNumber) {
		board->main_tmst_version = (uint8_t)json_value_get_number(val);
	}

	/* Optional: Configure board to enable/disable fine timestamp debug mode in DSP */
	val = json_object_get_value(root, "fts_debug");
	if (json_value_get_type(val) == JSONBoolean) {
		board->debug_tmst = (bool)json_value_get_boolean(val);
	}

	/* Get the number of chip on the board */
	sx1301_array = json_object_get_array(root, "SX1301_conf");
	if (sx1301_array != NULL) {
		board->nb_chip = (uint8_t)json_array_get_count(sx1301_array);
	}

	/* Check if spidev path is defined in JSON */
	val = json_object_get_value(root, "spidev");
	if (json_value_get_type(val) == JSONString) {
		spi_path = json_value_get_string(val);
		/* Open SPI link */
		x = spi_linuxdev_open(spi_path, -1, &spi_fd[nb_board]);
		if (x != 0) {
			error("Opening SPI failed, returned %i on open(%s)", x, spi_path);
			return -1;
		}
	}

	/* Configure SPI devices to master/slave modes (Needed for ATMEL CPU only)
     SPI 0: HOST <-> FPGA
     SPI 1: HOST/DSP <-> Flash
     */
	x = spi_set_mode(0, SPI_MODE_MASTER);
	if (x != 0) {
		error("Setting SPI mode failed, returned %i", x);
		return -1;
	}
	x = spi_set_mode(1, SPI_MODE_SLAVE); /* So that DSP can access the flash for booting */
	if (x != 0) {
		error("Setting SPI mode failed, returned %i", x);
		return -1;
	}

	/* Associate with SPI */
	switch (nb_board) {
	case 0:
		board->spi_read = &spi1_read;
		board->spi_write = &spi1_write;
		break;
#if (SX1301AR_MAX_BOARD_NB > 1)
	case 1:
		board->spi_read = &spi2_read;
		board->spi_write = &spi2_write;
		break;
#endif
#if (SX1301AR_MAX_BOARD_NB > 2)
	case 2:
		board->spi_read = &spi3_read;
		board->spi_write = &spi3_write;
		break;
#endif
#if (SX1301AR_MAX_BOARD_NB > 3)
	case 3:
		board->spi_read = &spi4_read;
		board->spi_write = &spi4_write;
		break;
#endif
	default:
		error("Board %u not supported", nb_board);
		return -1;
		break;
	}

	/* SX1301 chip configuration */
	if (sx1301_array == NULL) {
		return 0;
	}

	/* Chips configuration loop */
	for (nb_chip = 0;
	     nb_chip < json_array_get_count(sx1301_array) &&
	     nb_chip < SX1301AR_BOARD_CHIPS_NB;
	     ++nb_chip) {
		sx1301ar_chip_cfg_t *chip = &config_core->boards[nb_board].chips[nb_chip].chip;

		/* Point to chip configuration */
		chip_obj = json_array_get_object(sx1301_array, nb_chip);

		chip->enable = (bool)json_object_get_boolean(chip_obj, "chip_enable");
		if (chip->enable == false) {
			info("Chip %zd disabled", nb_chip);
			continue;
		}

		/* Configure enbaled chip */
		info("Chip %zd enabled", nb_chip);
		/* Mandatory: center frequency of the chip (ie. RX freq of each SX1301) */
		val = json_object_get_value(chip_obj, "chip_center_freq");
		if (json_value_get_type(val) == JSONNumber) {
			chip->freq_hz = (uint32_t)json_value_get_number(val);
		} else if (chip->freq_hz == 0) {
			error("Failed to parse mandatory parameter chip_center_freq for chip %zd", nb_chip);
			return -1;
		}

		// Rx RF chain select
		val = json_object_get_value(chip_obj, "chip_rf_chain");
		if (json_value_get_type(val) == JSONNumber) {
			chip->rf_chain = (uint8_t)json_value_get_number(val);
		} else if (chip->rf_chain == 0) {
			error("Failed to parse mandatory parameter chip_rf_chain for chip %zd", nb_chip);
			return -1;
		}

		/* Channels configuration loop */
		for (nb_chan = 0; nb_chan < SX1301AR_CHIP_CHAN_NB; ++nb_chan) {
			sx1301ar_chan_cfg_t *chan = &config_core->boards[nb_board].chips[nb_chip].chans[nb_chan];
			idx_chan = (nb_chip << 4) + nb_chan; /* generate board-level channel number */

			/* Point to channel configuration */
			if (nb_chan < SX1301AR_CHIP_MULTI_NB) {
				sprintf(param_name, "chan_multiSF_%zd", nb_chan);
				chan_obj = json_object_get_object(chip_obj, param_name);
			} else if (nb_chan == SX1301AR_CHIP_LSA_IDX) {
				chan_obj = json_object_get_object(chip_obj, "chan_LoRa_std");
			} else if (nb_chan == SX1301AR_CHIP_FSK_IDX) {
				chan_obj = json_object_get_object(chip_obj, "chan_FSK");
			}

			/* Channel disabled if configuration is absent */
			if (chan_obj == NULL) {
				chan->enable = false; /* should already be default value */
				info("Channel 0x%02X disabled", idx_chan);
				continue;
			} else {
				chan->enable = true;
			}

			JSON_PARSE_BOOLEAN(chan_obj, "enable", chan->enable);

			if (chan->enable == false) {
				continue;
			}

			/* Mandatory: center frequency of the channel */
			val = json_object_get_value(chan_obj, "chan_rx_freq");
			if (json_value_get_type(val) == JSONNumber) {
				chan->freq_hz = (uint32_t)json_value_get_number(val);
			} else if (chan->freq_hz == 0) {
				error("Failed to parse mandatory parameter chan_rx_freq for chip 0x%02X", idx_chan);
				return -1;
			}

			/* Parse parameters specific to each type of channel */
			if (nb_chan < SX1301AR_CHIP_MULTI_NB) {
				chan->bandwidth = BW_125K; /* not really necessary */

				/* LoRa spreading factor or spreading factor range (multi-SF modem) */
				val = json_object_get_value(chan_obj, "spread_factor");
				if (json_value_get_type(val) == JSONString) {
					j = sscanf(json_value_get_string(val), "%i-%i", &sf_a, &sf_b);
					if (j == 2) {
						chan->modrate = sx1301ar_sf_range_nb2enum(sf_a, sf_b);
					} else if (j == 1) {
						chan->modrate = sx1301ar_sf_nb2enum(sf_a);
					} else {
						chan->modrate = MR_UNDEFINED;
					}
				} else if (json_value_get_type(val) == JSONNumber) {
					chan->modrate = sx1301ar_sf_nb2enum((int)json_value_get_number(val));
				} else {
					chan->modrate = MR_UNDEFINED;
				}
			} else if (nb_chan == SX1301AR_CHIP_LSA_IDX) {
				/* Bandwidth */
				tmp_u32 = (uint32_t)json_object_get_number(chan_obj, "bandwidth");
				chan->bandwidth = sx1301ar_bw_nb2enum(tmp_u32);

				/* SF */
				tmp_u32 = (uint32_t)json_object_get_number(chan_obj, "spread_factor");
				chan->modrate = sx1301ar_sf_nb2enum(tmp_u32);
			} else if (nb_chan == SX1301AR_CHIP_FSK_IDX) {
				/* Bandwidth */
				tmp_u32 = (uint32_t)json_object_get_number(chan_obj, "bandwidth");
				chan->bandwidth = sx1301ar_bw_nb2enum(tmp_u32);

				/* Bit rate */
				tmp_u32 = (uint32_t)json_object_get_number(chan_obj, "bit_rate");
				chan->modrate = tmp_u32;
			}

			/* Commit channel configuration */
			if (nb_chan < SX1301AR_CHIP_MULTI_NB) {
				info("LoRa multi-SF channel 0x%02x configured at %u Hz, BW 125kHz, SF %i to %i",
				     idx_chan,
				     chan->freq_hz,
				     sx1301ar_sf_min_enum2nb(chan->modrate),
				     sx1301ar_sf_max_enum2nb(chan->modrate));

			} else if (nb_chan == SX1301AR_CHIP_LSA_IDX) {
				info("LoRa stand-alone channel 0x%02x configured at %u Hz, BW %li Hz, SF %i",
				     idx_chan,
				     chan->freq_hz,
				     sx1301ar_bw_enum2nb(chan->bandwidth),
				     sx1301ar_sf_enum2nb(chan->modrate));

			} else if (nb_chan == SX1301AR_CHIP_FSK_IDX) {
				info("FSK channel 0x%02x configured at %u Hz, BW %li Hz, bit rate %li",
				     idx_chan,
				     chan->freq_hz,
				     sx1301ar_bw_enum2nb(chan->bandwidth),
				     (long)chan->modrate);
			}
		}
	}

	return 0;
}

static int parse_sx1301(struct core *core, JSON_Object *root)
{
	struct config_core *config_core = core->config->core;
	JSON_Object *object = NULL;
	JSON_Array *array = NULL;
	size_t nb_board;

	if ((array = json_object_get_array(root, CONFIG_SX1301AR_JSONOBJ)) == NULL) {
		return 0;
	}

	for (nb_board = 0;
	     nb_board < json_array_get_count(array) &&
	     nb_board < SX1301AR_MAX_BOARD_NB;
	     nb_board++) {

		if ((object = json_array_get_object(array, nb_board)) == NULL) {
			return -1;

		} else if (parse_sx1301_single(core, object, nb_board) < 0) {
			return -1;

		} else if (config_core->nb_boards < nb_board + 1) {
			config_core->nb_boards = nb_board + 1;
		}
	}

	return 0;
}

static int parse_gateway(struct core *core, JSON_Object *root)
{
	struct config_beacon *beacon = &core->config->core->beacon;
	JSON_Object *object = NULL;
	int sf = 0, bw = 0;

	if ((object = json_object_get_object(root, "gateway_conf")) == NULL) {
		info("No gateway configuration");
		return 0;
	}

	JSON_PARSE_BOOLEAN(object, "beacon_enable", beacon->enable);
	JSON_PARSE_NUMBER(object, "beacon_period", beacon->period);
	JSON_PARSE_NUMBER(object, "beacon_freq_hz", beacon->frequency);
	JSON_PARSE_NUMBER(object, "beacon_freq_nb", beacon->frequencies);
	JSON_PARSE_NUMBER(object, "beacon_freq_step", beacon->frequency_step);
	JSON_PARSE_NUMBER(object, "beacon_datarate", sf);
	JSON_PARSE_NUMBER(object, "beacon_bw_hz", bw);
	JSON_PARSE_NUMBER(object, "beacon_power", beacon->power);
	JSON_PARSE_NUMBER(object, "beacon_infodesc", beacon->infodesc);
	JSON_PARSE_NUMBER(object, "ref_latitude", beacon->latitude);
	JSON_PARSE_NUMBER(object, "ref_longitude", beacon->longitude);

	if (bw) {
		switch (bw) {
		case 125000:
			beacon->bandwidth = BW_125K;
			break;
		case 500000:
			beacon->bandwidth = BW_500K;
			break;
		default:
			error("Invalid beacon bandwidth: %i", bw);
			return -1;
		}
	}

	if (sf) {
		switch (sf) {
		case 8:
			beacon->spreading_factor = MR_SF8;
			break;
		case 9:
			beacon->spreading_factor = MR_SF9;
			break;
		case 10:
			beacon->spreading_factor = MR_SF10;
			break;
		case 12:
			beacon->spreading_factor = MR_SF12;
			break;
		default:
			error("Invalid beacon spreading factor: %i", sf);
			return -1;
		}
	}

	if (beacon->period > 0 && beacon->period < 6) {
		error("Invalid configuration for Beacon period, must be >= 6s");
		return -1;
	}

	return 0;
}

struct config_core *config_core_init(void)
{
	struct config_core *config_core = NULL;
	size_t nb_board, nb_chip, nb_chan;

	if ((config_core = zalloc(sizeof(*config_core))) == NULL) {
		eerror("Failed to allocate core configuration");
		return NULL;
	}

	for (nb_board = 0; nb_board < SX1301AR_MAX_BOARD_NB; nb_board++) {
		spi_fd[nb_board] = -1;
		pthread_mutex_init(&spi_mutex[nb_board], NULL);

		struct config_board *board = &config_core->boards[nb_board];
		board->board = sx1301ar_init_board_cfg();
		board->lbt = sx1301ar_init_lbt_cfg();

		for (nb_chip = 0; nb_chip < SX1301AR_BOARD_CHIPS_NB; nb_chip++) {
			struct config_chip *chip = &board->chips[nb_chip];
			chip->chip = sx1301ar_init_chip_cfg();

			for (nb_chan = 0; nb_chan < SX1301AR_CHIP_CHAN_NB; nb_chan++) {
				chip->chans[nb_chan] = sx1301ar_init_chan_cfg();
			}
		}
	}

	config_core->beacon.frequency = 869525000;
	config_core->beacon.frequencies = 1;
	config_core->beacon.bandwidth = BW_125K;
	config_core->beacon.spreading_factor = MR_SF9;
	config_core->beacon.power = 14;

	return config_core;
}

int config_core_parse(struct core *core, JSON_Object *root)
{
	if (parse_sx1301(core, root) < 0) {
		error("Failed to parse SX1301 configuration");
		return -1;

	} else if (parse_gateway(core, root) < 0) {
		error("Failed to parse gateway configuration");
		return -1;
	}

	return 0;
}

int config_core_apply(struct core *core)
{
	struct config_core *config_core = core->config->core;
	size_t nb_board, nb_chip, nb_chan, nb_rfchain;

	for (nb_board = 0; nb_board < config_core->nb_boards; nb_board++) {
		struct config_board *board = &config_core->boards[nb_board];

		/* Compute RSSI offset taking into account antenna gain and insertion loss */
		for (nb_rfchain = 0;
		     nb_rfchain < SX1301AR_BOARD_RFCHAIN_NB;
		     nb_rfchain++) {
			board->board.rf_chain[nb_rfchain].rssi_offset = board->board.rf_chain[nb_rfchain].rssi_offset -
									board->board.rf_chain[nb_rfchain].antenna_gain + board->board.rf_chain[nb_rfchain].insertion_loss;
			info("rf_chain %zd: new RSSI offset including antenna gain (%f) and path loss (%f): %.1f\n",
			     nb_rfchain, board->board.rf_chain[nb_rfchain].antenna_gain, board->board.rf_chain[nb_rfchain].insertion_loss,
			     board->board.rf_chain[nb_rfchain].rssi_offset);
		}

		if (sx1301ar_conf_board(nb_board, &board->board) < 0) {
			lerror("Failed to configure board");
			return -1;

		} else if (sx1301ar_conf_lbt(nb_board, &board->lbt) < 0) {
			lerror("Failed to configure LBT");
			return -1;
		}

		for (nb_chip = 0; nb_chip < SX1301AR_BOARD_CHIPS_NB; nb_chip++) {
			struct config_chip *chip = &board->chips[nb_chip];

			if (sx1301ar_conf_chip(nb_board, nb_chip, &chip->chip) < 0) {
				lerror("Failed to configure chip %zu", nb_chip);
				return -1;
			}

			for (nb_chan = 0; nb_chan < SX1301AR_CHIP_CHAN_NB; nb_chan++) {
				uint8_t idx_chan = (nb_chip << 4) + nb_chan;

				if (sx1301ar_conf_chan(nb_board, idx_chan, &chip->chans[nb_chan]) < 0) {
					lerror("Failed to configure channel %u", idx_chan);
					return -1;
				}
			}
		}
	}

	return 0;
}

void config_core_free(struct core *core)
{
	size_t nb_board;

	if (core && core->config) {
		for (nb_board = 0; nb_board < core->config->core->nb_boards; nb_board++) {
			if (spi_linuxdev_close(spi_fd[nb_board]) < 0) {
				warning("Failed to close SPI");
			}
		}

		free(core->config->core);
	}
}
