/*
 / _____)             _              | |
( (____  _____ ____ _| |_ _____  ____| |__
 \____ \| ___ |    (_   _) ___ |/ ___)  _ \
 _____) ) ____| | | || |_| ____( (___| | | |
(______/|_____)_|_|_| \__)_____)\____)_| |_|
  (C)2013 Semtech-Cycleo

Description:
    Host specific functions to address the LoRa concentrator registers through
    a SPI interface.
    Single-byte read/write and burst read/write.
    Does not handle pagination.
    Could be used with multiple SPI ports in parallel (explicit file descriptor)

License: Revised BSD License, see LICENSE.TXT file include in the project
Maintainer: Sylvain Miermont
*/

/* -------------------------------------------------------------------------- */
/* --- DEPENDENCIES --------------------------------------------------------- */

#include <fcntl.h>  /* open */
#include <stdint.h> /* C99 types */
#include <stdio.h>  /* printf fprintf */
#include <stdlib.h> /* malloc free */
#include <string.h> /* memset */
#include <unistd.h> /* lseek, close */

#include <linux/spi/spidev.h>
#include <sys/ioctl.h>

#include "aux.h" /* wait_ms */
#include "hal.h"
#include "spi.h"

#include "utils/logger.h"

/* -------------------------------------------------------------------------- */
/* --- PRIVATE MACROS ------------------------------------------------------- */

#define ARRAY_SIZE(a) (sizeof(a) / sizeof((a)[0]))
#define CHECK_NULL(a)                 \
	if (a == NULL) {              \
		return LGW_SPI_ERROR; \
	}

/* -------------------------------------------------------------------------- */
/* --- PRIVATE CONSTANTS ---------------------------------------------------- */

#define READ_ACCESS 0x00
#define WRITE_ACCESS 0x80
#define SPI_SPEED 2000000
#define SPI_DEV_PATH "/dev/spidev0.0"
#define EEPROM_SPI_SPEED 1000000
#define FPGA_SPI_SPEED 1000000

/* Wirnet iFemtocell Architecture */
#if defined(PLATFORM_WIFC)
#define FPGA_SPI_DEV_PATH "/dev/spidev1.1"
#define FPGA_CSN_GPIONO 128
#define FPGA_CRESET_GPIONO 129
#define FPGA_CDONE_GPIONO 130
#define BUFFER_MAX_SIZE 64
#define GPIO_IN 0
#define GPIO_OUT 1
#define GPIO_LOW 0
#define GPIO_HIGH 1
#endif

/* Wirnet Station Architecture */
#if defined(PLATFORM_WSTA)
#undef SPI_DEV_PATH
#define SPI_DEV_PATH "/dev/spidev32766.0"
#define FPGA_SPI_DEV_PATH "/dev/spidev32766.1"
#define FPGA_CSN_DEV_PATH "/dev/p1_20"
#define FPGA_CRESET_DEV_PATH "/dev/p1_19"
#define FPGA_CDONE_DEV_PATH "/dev/p1_16"
#endif

/* -------------------------------------------------------------------------- */
/* --- PRIVATE VARIABLES ---------------------------------------------------- */

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
static int spi_fpga_fd = -1;
#endif

#if defined(PLATFORM_WSTA)
static int fpga_csn_fd = -1;
static int fpga_creset_fd = -1;
static int fpga_cdone_fd = -1;
#endif

/* -------------------------------------------------------------------------- */
/* --- PRIVATE FUNCTIONS DEFINITION ------------------------------------------ */

#if defined(PLATFORM_WIFC)
#include <sys/stat.h>

/* Use GPIO lib (sysfs User interface) */
static bool gpio_is_exported(int pin)
{
	char path[BUFFER_MAX_SIZE];
	struct stat info;

	snprintf(path, BUFFER_MAX_SIZE, "/sys/class/gpio/gpio%d", pin);
	if (stat(path, &info) == 0) {
		return true;
	}

	return false;
}

static int gpio_export(int pin)
{
	char buffer[BUFFER_MAX_SIZE];
	ssize_t bytes_written;
	int fd;

	/* Check first if it's not yet exported */
	if (gpio_is_exported(pin)) {
		return 0;
	}

	fd = open("/sys/class/gpio/export", O_WRONLY);
	if (fd == -1) {
		error("Failed to open export for writing");
		return -1;
	}

	bytes_written = snprintf(buffer, BUFFER_MAX_SIZE, "%d", pin);

	if (write(fd, buffer, bytes_written) == -1) {
		error("Failed to export gpio");
		close(fd);
		return -1;
	}

	return 0;
}

static int gpio_get_direction(int pin)
{
	char path[BUFFER_MAX_SIZE];
	char value_str[5];
	int fd;

	snprintf(path, BUFFER_MAX_SIZE, "/sys/class/gpio/gpio%d/direction", pin);
	fd = open(path, O_RDWR);
	if (fd == -1) {
		error("Failed to open gpio direction");
		return -1;
	}

	if (read(fd, value_str, sizeof(value_str)) == -1) {
		error("Failed to read value");
		close(fd);
		return -1;
	}

	close(fd);

	if (!strncmp(value_str, "out", 3)) {
		return GPIO_OUT;
	} else if (!strncmp(value_str, "in", 2)) {
		return GPIO_IN;
	}

	return -2;
}

static int gpio_set_direction(int pin, int dir)
{
	char path[BUFFER_MAX_SIZE];
	int fd;

	/* Check first */
	if (dir == gpio_get_direction(pin)) {
		return 0;
	}

	snprintf(path, BUFFER_MAX_SIZE, "/sys/class/gpio/gpio%d/direction", pin);
	fd = open(path, O_RDWR);
	if (fd == -1) {
		error("Failed to open gpio direction");
		return -1;
	}

	if (write(fd, GPIO_IN == dir ? "in" : "out", GPIO_IN == dir ? 2 : 3) == -1) {
		error("Failed to set direction");
		close(fd);
		return -1;
	}

	close(fd);
	return 0;
}

static int gpio_read(int pin, int *value)
{
	char path[BUFFER_MAX_SIZE];
	char value_str[2];
	int fd;

	snprintf(path, BUFFER_MAX_SIZE, "/sys/class/gpio/gpio%d/value", pin);
	fd = open(path, O_RDONLY);
	if (fd == -1) {
		error("Failed to open gpio value for reading");
		return -1;
	}

	if (read(fd, value_str, sizeof(value_str)) == -1) {
		error("Failed to read value");
		close(fd);
		return -1;
	}

	close(fd);
	*value = !!atoi(value_str);
	return 0;
}

static int gpio_write(int pin, int value)
{
	char path[BUFFER_MAX_SIZE];
	int fd;

	snprintf(path, BUFFER_MAX_SIZE, "/sys/class/gpio/gpio%d/value", pin);
	fd = open(path, O_WRONLY);
	if (fd == -1) {
		error("Failed to open gpio value for writing");
		return -1;
	}

	if (write(fd, value == GPIO_LOW ? "0" : "1", 1) != 1) {
		error("Failed to write value");
		close(fd);
		return -1;
	}

	close(fd);
	return 0;
}

static void lgw_fpga_set_csn(uint8_t level)
{
	int err = 0;

	if (!err)
		err = gpio_export(FPGA_CSN_GPIONO);
	if (!err)
		err = gpio_set_direction(FPGA_CSN_GPIONO, GPIO_OUT);
	if (!err)
		err = gpio_write(FPGA_CSN_GPIONO, level ? GPIO_HIGH : GPIO_LOW);
}

static void lgw_fpga_set_creset(uint8_t level)
{
	int err = 0;

	if (!err)
		err = gpio_export(FPGA_CRESET_GPIONO);
	if (!err)
		err = gpio_set_direction(FPGA_CRESET_GPIONO, GPIO_OUT);
	if (!err)
		err = gpio_write(FPGA_CRESET_GPIONO, level ? GPIO_HIGH : GPIO_LOW);
}

static int lgw_fpga_get_cdone(void)
{
	int value, err = 0;

	if (!err)
		err = gpio_export(FPGA_CDONE_GPIONO);
	if (!err)
		err = gpio_set_direction(FPGA_CDONE_GPIONO, GPIO_IN);
	if (!err)
		err = gpio_read(FPGA_CDONE_GPIONO, &value);

	if (err)
		value = -1;

	return value;
}
#endif

#if defined(PLATFORM_WSTA)
#include <linux/kerlink_gpio.h>

static void lgw_fpga_set_csn(uint8_t level)
{
	if (fpga_csn_fd < 0) {
		// Open it
		fpga_csn_fd = open(FPGA_CSN_DEV_PATH, O_RDWR | O_NONBLOCK);
		if (fpga_csn_fd < 0) {
			return;
		}
	}

	if (level == 0) {
		ioctl(fpga_csn_fd, GPIO_WRITE_LOW_LEVEL, NULL);
	} else {
		ioctl(fpga_csn_fd, GPIO_WRITE_HIGH_LEVEL, NULL);
	}
}

static void lgw_fpga_set_creset(uint8_t level)
{
	if (fpga_creset_fd < 0) {
		// Open it
		fpga_creset_fd = open(FPGA_CRESET_DEV_PATH, O_RDWR | O_NONBLOCK);
		if (fpga_creset_fd < 0) {
			return;
		}
	}

	if (level == 0) {
		ioctl(fpga_creset_fd, GPIO_WRITE_LOW_LEVEL, NULL);
	} else {
		ioctl(fpga_creset_fd, GPIO_WRITE_HIGH_LEVEL, NULL);
	}
}

static int lgw_fpga_get_cdone(void)
{
	int ret, readval;

	if (fpga_cdone_fd < 0) {
		// Open it
		fpga_cdone_fd = open(FPGA_CDONE_DEV_PATH, O_RDWR | O_NONBLOCK);
		if (fpga_csn_fd < 0) {
			return -1;
		}
	}

	ret = ioctl(fpga_cdone_fd, GPIO_READ, &readval);

	if (ret < 0) {
		return -1;
	} else {
		return readval;
	}
}
#endif

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
static int lgw_fpga_spi_open(void)
{
	int dev;
	int a = 0, b = 0;
	int i;

	/* open SPI device */
	dev = open(FPGA_SPI_DEV_PATH, O_RDWR);
	if (dev < 0) {
		error("failed to open SPI FPGA device %s", FPGA_SPI_DEV_PATH);
		return LGW_SPI_ERROR;
	}

	/* setting SPI mode to 'mode 0' */
	i = SPI_MODE_0;
	a = ioctl(dev, SPI_IOC_WR_MODE, &i);
	b = ioctl(dev, SPI_IOC_RD_MODE, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI FPGA PORT FAIL TO SET IN MODE 0");
		close(dev);
		return LGW_SPI_ERROR;
	}

	/* setting SPI max clk (in Hz) */
	i = FPGA_SPI_SPEED;
	a = ioctl(dev, SPI_IOC_WR_MAX_SPEED_HZ, &i);
	b = ioctl(dev, SPI_IOC_RD_MAX_SPEED_HZ, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI FPGA PORT FAIL TO SET MAX SPEED");
		close(dev);
		return LGW_SPI_ERROR;
	}

	/* setting SPI to MSB first */
	i = 0;
	a = ioctl(dev, SPI_IOC_WR_LSB_FIRST, &i);
	b = ioctl(dev, SPI_IOC_RD_LSB_FIRST, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI FPGA PORT FAIL TO SET MSB FIRST");
		close(dev);
		return LGW_SPI_ERROR;
	}

	/* setting SPI to 8 bits per word */
	i = 0;
	a = ioctl(dev, SPI_IOC_WR_BITS_PER_WORD, &i);
	b = ioctl(dev, SPI_IOC_RD_BITS_PER_WORD, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI FPGA PORT FAIL TO SET 8 BITS-PER-WORD");
		close(dev);
		return LGW_SPI_ERROR;
	}

	spi_fpga_fd = dev;
	debug("SPI FPGA port opened and configured ok");
	return LGW_SPI_SUCCESS;
}

static int lgw_fpga_change_spimode(unsigned char mode)
{
	int a = 0, b = 0;
	int i;

	/* setting SPI mode to 'mode 0' */
	i = mode;
	a = ioctl(spi_fpga_fd, SPI_IOC_WR_MODE, &i);
	b = ioctl(spi_fpga_fd, SPI_IOC_RD_MODE, &i);

	if ((a < 0) || (b < 0)) {
		error("SPI PORT FAIL TO SET IN MODE %u", mode);
		return LGW_SPI_ERROR;
	}

	return LGW_SPI_SUCCESS;
}

int lgw_fpga_program(uint8_t *data, uint16_t size)
{
	struct spi_ioc_transfer k[32];
	int size_to_do, chunk_size, offset;
	int byte_transfered = 0;
	int i, ret;
	uint8_t dummy_byte[12] = {
	    0x00, 0xFF, 0x00, 0xFF,
	    0x00, 0xFF, 0x00, 0xFF,
	    0x00, 0xFF, 0x00, 0xFF};

	CHECK_NULL(data);
	if (size == 0) {
		error("BURST OF NULL LENGTH");
		return LGW_SPI_ERROR;
	}

	ret = lgw_fpga_spi_open();
	if (ret != LGW_SPI_SUCCESS) {
		return ret;
	}

	/* Change Mode for FPGA programming */
	lgw_fpga_change_spimode(SPI_MODE_3);

	/*Set HOST_CSN_FPGA High*/
	lgw_fpga_set_csn(1);
	wait_ms(100);
	/*Set CRESET Low*/
	lgw_fpga_set_creset(0);
	/*Set HOST_CSN_FPGA Low*/
	lgw_fpga_set_csn(0);
	/* Test programmation with SPI transaction */
	//lgw_spi_wb(spi_fpga_fd, LGW_SPI_MUX_MODE0, LGW_SPI_MUX_TARGET_FPGA, 0, );
	k[0].tx_buf = (unsigned long)(dummy_byte);
	k[0].len = 12;
	k[0].cs_change = 1;
	byte_transfered += ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k);
	//debug("BURST WRITE: transferred %d ", byte_transfered);
	/* */
	wait_ms(10);
	/*Set CRESET High*/
	lgw_fpga_set_creset(1);
	wait_ms(10);
	/*Set HOST_CSN_FPGA High*/
	lgw_fpga_set_csn(1);
	wait_ms(10);

	/*Send FPGA programmation file*/
	size_to_do = size;
	memset(&k, 0, sizeof(k)); /* clear k */
	byte_transfered = 0;
	for (i = 0; size_to_do > 0; ++i) {
		chunk_size = (size_to_do < LGW_BURST_CHUNK) ? size_to_do : LGW_BURST_CHUNK;
		offset = i * LGW_BURST_CHUNK;
		k[0].tx_buf = (unsigned long)(data + offset);
		k[0].len = chunk_size;
		k[0].cs_change = 1;
		size_to_do -= chunk_size; /* subtract the quantity of data already transferred */

		/*Set HOST_CSN_FPGA Low*/
		lgw_fpga_set_csn(0);
		wait_ms(1);
		/*Send SPI chunked burst*/
		byte_transfered += (ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k));
		wait_ms(1);
		/*Set HOST_CSN_FPGA High*/
		lgw_fpga_set_csn(1);
		wait_ms(1);
		//debug("MESSAGE #%d = len : %d ",i+1,chunk_size);
	}

	/*Set HOST_CSN_FPGA Low*/
	lgw_fpga_set_csn(0);

	/*Send FPGA programmation dummy bit > 49*/
	k[0].tx_buf = (unsigned long)(dummy_byte);
	k[0].len = 12;
	k[0].cs_change = 1;
	byte_transfered += ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k);
	//debug("BURST WRITE: transferred %d ", byte_transfered);

	/*Set HOST_CSN_FPGA High*/
	lgw_fpga_set_csn(1);
	wait_ms(10);

	/* Change Mode now FPGA is programmed */
	lgw_fpga_change_spimode(SPI_MODE_0);

	/* Finally, close device */
	close(spi_fpga_fd);
	spi_fpga_fd = -1;

	memset(&k, 0, sizeof(k)); /* clear k */
	/* determine return code */
	if (byte_transfered != (size + 12)) {
		error("SPI BURST WRITE FAILURE");
		return LGW_SPI_ERROR;
	} else {
		/* Check CDONE pin */
		ret = lgw_fpga_get_cdone();
		if (ret == 1) {
			debug("program FPGA done CDONE TEST PASSED");
		} else {
			error("program FPGA failed CDONE TEST FAILED");
			return LGW_SPI_ERROR;
		}

		debug("success connecting and programming the FPGA concentrator");

		//debug("SPI burst write success");
		return LGW_SPI_SUCCESS;
	}
}
#endif

/* -------------------------------------------------------------------------- */
/* --- PUBLIC FUNCTIONS DEFINITION ------------------------------------------ */

/* SPI initialization and configuration */
int lgw_spi_open(void **spi_target_ptr)
{
	int *spi_device = NULL;
	int dev;
	int a = 0, b = 0;
	int i;

	/* check input variables */
	CHECK_NULL(spi_target_ptr); /* cannot be null, must point on a void pointer (*spi_target_ptr can be null) */

	/* allocate memory for the device descriptor */
	spi_device = malloc(sizeof(int));
	if (spi_device == NULL) {
		error("MALLOC FAIL");
		return LGW_SPI_ERROR;
	}

	/* open SPI device */
	dev = open(SPI_DEV_PATH, O_RDWR);
	if (dev < 0) {
		error("failed to open SPI device %s", SPI_DEV_PATH);
		return LGW_SPI_ERROR;
	}

	/* setting SPI mode to 'mode 0' */
	i = SPI_MODE_0;
	a = ioctl(dev, SPI_IOC_WR_MODE, &i);
	b = ioctl(dev, SPI_IOC_RD_MODE, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI PORT FAIL TO SET IN MODE 0");
		close(dev);
		free(spi_device);
		return LGW_SPI_ERROR;
	}

	/* setting SPI max clk (in Hz) */
	i = SPI_SPEED;
	a = ioctl(dev, SPI_IOC_WR_MAX_SPEED_HZ, &i);
	b = ioctl(dev, SPI_IOC_RD_MAX_SPEED_HZ, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI PORT FAIL TO SET MAX SPEED");
		close(dev);
		free(spi_device);
		return LGW_SPI_ERROR;
	}

	/* setting SPI to MSB first */
	i = 0;
	a = ioctl(dev, SPI_IOC_WR_LSB_FIRST, &i);
	b = ioctl(dev, SPI_IOC_RD_LSB_FIRST, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI PORT FAIL TO SET MSB FIRST");
		close(dev);
		free(spi_device);
		return LGW_SPI_ERROR;
	}

	/* setting SPI to 8 bits per word */
	i = 0;
	a = ioctl(dev, SPI_IOC_WR_BITS_PER_WORD, &i);
	b = ioctl(dev, SPI_IOC_RD_BITS_PER_WORD, &i);
	if ((a < 0) || (b < 0)) {
		error("SPI PORT FAIL TO SET 8 BITS-PER-WORD");
		close(dev);
		return LGW_SPI_ERROR;
	}

	*spi_device = dev;
	*spi_target_ptr = (void *)spi_device;
	debug("SPI port opened and configured ok");

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	return lgw_fpga_spi_open();
#else
	return LGW_SPI_SUCCESS;
#endif
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* SPI release */
int lgw_spi_close(void *spi_target)
{
	int spi_device;
	int a;

	/* check input variables */
	CHECK_NULL(spi_target);

	/* close file & deallocate file descriptor */
	spi_device = *(int *)spi_target; /* must check that spi_target is not null beforehand */
	a = close(spi_device);
	free(spi_target);

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	close(spi_fpga_fd);
#endif

	/* determine return code */
	if (a < 0) {
		error("SPI PORT FAILED TO CLOSE");
		return LGW_SPI_ERROR;
	} else {
		debug("SPI port closed");
		return LGW_SPI_SUCCESS;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* Simple write */
int lgw_spi_w(void *spi_target, uint8_t spi_mux_mode, uint8_t spi_mux_target, uint8_t address, uint8_t data)
{
	int spi_device;
	uint8_t out_buf[3];
	uint8_t command_size;
	struct spi_ioc_transfer k;
	int a;

	/* check input variables */
	CHECK_NULL(spi_target);
	if ((address & 0x80) != 0) {
		warning("SPI address > 127");
	}

	spi_device = *(int *)spi_target; /* must check that spi_target is not null beforehand */

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
		spi_mux_mode = LGW_SPI_MUX_MODE0;
	}
#endif

	/* prepare frame to be sent */
	if (spi_mux_mode == LGW_SPI_MUX_MODE1) {
		out_buf[0] = spi_mux_target;
		out_buf[1] = WRITE_ACCESS | (address & 0x7F);
		out_buf[2] = data;
		command_size = 3;
	} else {
		out_buf[0] = WRITE_ACCESS | (address & 0x7F);
		out_buf[1] = data;
		command_size = 2;
	}

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k.tx_buf = (unsigned long)out_buf;
	k.len = command_size;
	k.speed_hz = SPI_SPEED;
	k.cs_change = 0;
	k.bits_per_word = 8;
#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
		a = ioctl(spi_device, SPI_IOC_MESSAGE(1), &k);
	} else {
		lgw_fpga_set_csn(0);
		a = ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k);
		lgw_fpga_set_csn(1);
	}
#else
	a = ioctl(spi_device, SPI_IOC_MESSAGE(1), &k);
#endif

	/* determine return code */
	if (a != (int)k.len) {
		error("SPI WRITE FAILURE");
		return LGW_SPI_ERROR;
	} else {
		//debug("SPI write success");
		return LGW_SPI_SUCCESS;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* Simple read */
int lgw_spi_r(void *spi_target, uint8_t spi_mux_mode, uint8_t spi_mux_target, uint8_t address, uint8_t *data)
{
	int spi_device;
	uint8_t out_buf[3];
	uint8_t command_size;
	uint8_t in_buf[ARRAY_SIZE(out_buf)];
	struct spi_ioc_transfer k;
	int a;

	/* check input variables */
	CHECK_NULL(spi_target);
	if ((address & 0x80) != 0) {
		warning("SPI address > 127");
	}
	CHECK_NULL(data);

	spi_device = *(int *)spi_target; /* must check that spi_target is not null beforehand */

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
		spi_mux_mode = LGW_SPI_MUX_MODE0;
	}
#endif

	/* prepare frame to be sent */
	if (spi_mux_mode == LGW_SPI_MUX_MODE1) {
		out_buf[0] = spi_mux_target;
		out_buf[1] = READ_ACCESS | (address & 0x7F);
		out_buf[2] = 0x00;
		command_size = 3;
	} else {
		out_buf[0] = READ_ACCESS | (address & 0x7F);
		out_buf[1] = 0x00;
		command_size = 2;
	}

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k.tx_buf = (unsigned long)out_buf;
	k.rx_buf = (unsigned long)in_buf;
	k.len = command_size;
	k.cs_change = 0;
#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
		a = ioctl(spi_device, SPI_IOC_MESSAGE(1), &k);
	} else {
		lgw_fpga_set_csn(0);
		a = ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k);
		lgw_fpga_set_csn(1);
	}
#else
	a = ioctl(spi_device, SPI_IOC_MESSAGE(1), &k);
#endif

	/* determine return code */
	if (a != (int)k.len) {
		error("SPI READ FAILURE");
		return LGW_SPI_ERROR;
	} else {
		//debug("SPI read success");
		*data = in_buf[command_size - 1];
		return LGW_SPI_SUCCESS;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* Burst (multiple-byte) write */
int lgw_spi_wb(void *spi_target, uint8_t spi_mux_mode, uint8_t spi_mux_target, uint8_t address, uint8_t *data, uint16_t size)
{
	int spi_device;
	uint8_t command[2];
	uint8_t command_size;
	struct spi_ioc_transfer k[2];
	int size_to_do, chunk_size, offset;
	int byte_transfered = 0;
	int i;

	/* check input parameters */
	CHECK_NULL(spi_target);
	if ((address & 0x80) != 0) {
		warning("SPI address > 127");
	}
	CHECK_NULL(data);
	if (size == 0) {
		error("BURST OF NULL LENGTH");
		return LGW_SPI_ERROR;
	}

	spi_device = *(int *)spi_target; /* must check that spi_target is not null beforehand */

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
		spi_mux_mode = LGW_SPI_MUX_MODE0;
	}
#endif

	/* prepare command byte */
	if (spi_mux_mode == LGW_SPI_MUX_MODE1) {
		command[0] = spi_mux_target;
		command[1] = WRITE_ACCESS | (address & 0x7F);
		command_size = 2;
	} else {
		command[0] = WRITE_ACCESS | (address & 0x7F);
		command_size = 1;
	}
	size_to_do = size;

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k[0].tx_buf = (unsigned long)&command[0];
	k[0].len = command_size;
	k[0].cs_change = 0;
	k[1].cs_change = 0;
	for (i = 0; size_to_do > 0; ++i) {
		chunk_size = (size_to_do < LGW_BURST_CHUNK) ? size_to_do : LGW_BURST_CHUNK;
		offset = i * LGW_BURST_CHUNK;
		k[1].tx_buf = (unsigned long)(data + offset);
		k[1].len = chunk_size;
#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
		if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
			byte_transfered += (ioctl(spi_device, SPI_IOC_MESSAGE(2), &k) - k[0].len);
		} else {
			lgw_fpga_set_csn(0);
			byte_transfered += (ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(2), &k) - k[0].len);
			lgw_fpga_set_csn(1);
		}
#else
		byte_transfered += (ioctl(spi_device, SPI_IOC_MESSAGE(2), &k) - k[0].len);
#endif
		//debug("BURST WRITE: to trans %d # chunk %d # transferred %d ", size_to_do, chunk_size, byte_transfered);
		size_to_do -= chunk_size; /* subtract the quantity of data already transferred */
	}

	/* determine return code */
	if (byte_transfered != size) {
		error("SPI BURST WRITE FAILURE");
		return LGW_SPI_ERROR;
	} else {
		//debug("SPI burst write success");
		return LGW_SPI_SUCCESS;
	}
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/* Burst (multiple-byte) read */
int lgw_spi_rb(void *spi_target, uint8_t spi_mux_mode, uint8_t spi_mux_target, uint8_t address, uint8_t *data, uint16_t size)
{
	int spi_device;
	uint8_t command[2];
	uint8_t command_size;
	struct spi_ioc_transfer k[2];
	int size_to_do, chunk_size, offset;
	int byte_transfered = 0;
	int i;

	/* check input parameters */
	CHECK_NULL(spi_target);
	if ((address & 0x80) != 0) {
		warning("SPI address > 127");
	}
	CHECK_NULL(data);
	if (size == 0) {
		error("BURST OF NULL LENGTH");
		return LGW_SPI_ERROR;
	}

	spi_device = *(int *)spi_target; /* must check that spi_target is not null beforehand */

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
	if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
		spi_mux_mode = LGW_SPI_MUX_MODE0;
	}
#endif

	/* prepare command byte */
	if (spi_mux_mode == LGW_SPI_MUX_MODE1) {
		command[0] = spi_mux_target;
		command[1] = READ_ACCESS | (address & 0x7F);
		command_size = 2;
	} else {
		command[0] = READ_ACCESS | (address & 0x7F);
		command_size = 1;
	}
	size_to_do = size;

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k[0].tx_buf = (unsigned long)&command[0];
	k[0].len = command_size;
	k[0].cs_change = 0;
	k[1].cs_change = 0;
	for (i = 0; size_to_do > 0; ++i) {
		chunk_size = (size_to_do < LGW_BURST_CHUNK) ? size_to_do : LGW_BURST_CHUNK;
		offset = i * LGW_BURST_CHUNK;
		k[1].rx_buf = (unsigned long)(data + offset);
		k[1].len = chunk_size;
#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
		if (spi_mux_target == LGW_SPI_MUX_TARGET_SX1301) {
			byte_transfered += (ioctl(spi_device, SPI_IOC_MESSAGE(2), &k) - k[0].len);
		} else {
			lgw_fpga_set_csn(0);
			byte_transfered += (ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(2), &k) - k[0].len);
			lgw_fpga_set_csn(1);
		}
#else
		byte_transfered += (ioctl(spi_device, SPI_IOC_MESSAGE(2), &k) - k[0].len);
#endif
		//debug("BURST READ: to trans %d # chunk %d # transferred %d ", size_to_do, chunk_size, byte_transfered);
		size_to_do -= chunk_size; /* subtract the quantity of data already transferred */
	}

	/* determine return code */
	if (byte_transfered != size) {
		error("SPI BURST READ FAILURE");
		return LGW_SPI_ERROR;
	} else {
		//debug("SPI burst read success");
		return LGW_SPI_SUCCESS;
	}
}

#if defined(PLATFORM_WIFC) || defined(PLATFORM_WSTA)
int lgw_spi_eeprom_read_word(uint8_t address, uint16_t *data)
{
	uint8_t out_buf0[3];
	uint8_t out_buf1[2];
	uint8_t in_buf0[ARRAY_SIZE(out_buf0)];
	uint8_t in_buf1[ARRAY_SIZE(out_buf1)];
	struct spi_ioc_transfer k[2];
	unsigned int byte_transfered = 0;
	int a0, a1;
	int i;

	/* prepare frame to be sent */
	out_buf0[0] = LGW_SPI_MUX_TARGET_EEPROM;
	out_buf0[1] = 0x01;
	out_buf0[2] = 0x80 | (address & 0x3F);
	out_buf1[0] = 0x00;
	out_buf1[1] = 0x00;

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k[0].tx_buf = (unsigned long)out_buf0;
	k[0].rx_buf = (unsigned long)in_buf0;
	k[0].len = ARRAY_SIZE(out_buf0);
	k[0].speed_hz = EEPROM_SPI_SPEED;
	k[0].delay_usecs = 10;
	k[0].cs_change = 0;
	k[0].bits_per_word = 8;

	k[1].tx_buf = (unsigned long)out_buf1;
	k[1].rx_buf = (unsigned long)in_buf1;
	k[1].len = ARRAY_SIZE(out_buf1);
	k[1].speed_hz = EEPROM_SPI_SPEED;
	k[1].delay_usecs = 2;
	k[1].cs_change = 0;
	k[1].bits_per_word = 8;

	/*Set HOST_CSN_FPGA Low*/
	lgw_fpga_set_csn(0);
	/*Send SPI transaction*/
	a0 = ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k[0]);

	i = SPI_MODE_1;
	ioctl(spi_fpga_fd, SPI_IOC_WR_MODE, &i);
	ioctl(spi_fpga_fd, SPI_IOC_RD_MODE, &i);
	a1 = ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k[1]);
	byte_transfered = a0 + a1;
	/*Set HOST_CSN_FPGA High*/
	lgw_fpga_set_csn(1);

	/* go back to original spi_mode */

	i = SPI_MODE_0;
	ioctl(spi_fpga_fd, SPI_IOC_WR_MODE, &i);
	ioctl(spi_fpga_fd, SPI_IOC_RD_MODE, &i);
	/* determine return code */
	if (byte_transfered != (k[0].len + k[1].len)) {
		error("SPI READ FAILURE (byte_transfered = %d)", byte_transfered);
		return LGW_SPI_ERROR;
	} else {
		//debug("EEPROM read success");
		*data = (in_buf1[0] << 8) + in_buf1[1];
		return LGW_SPI_SUCCESS;
	}
}

int lgw_spi_eeprom_write_word(uint8_t address, uint16_t data)
{
	uint8_t out_buf0[5];
	uint8_t in_buf0[ARRAY_SIZE(out_buf0)];
	struct spi_ioc_transfer k;
	int byte_transfered = 0;

	/* prepare frame to be sent */
	out_buf0[0] = LGW_SPI_MUX_TARGET_EEPROM;
	out_buf0[1] = 0x01;
	out_buf0[2] = 0x40 | (address & 0x3F);
	out_buf0[3] = (data >> 8) & 0xFF;
	out_buf0[4] = (data & 0xFF);

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k.tx_buf = (unsigned long)out_buf0;
	k.rx_buf = (unsigned long)in_buf0;
	k.len = ARRAY_SIZE(out_buf0);
	k.speed_hz = EEPROM_SPI_SPEED;
	k.delay_usecs = 2;
	k.cs_change = 0;
	k.bits_per_word = 8;

	/*Set HOST_CSN_FPGA Low*/
	lgw_fpga_set_csn(0);
	/*Send SPI transaction*/
	byte_transfered = ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k);

	/*Set HOST_CSN_FPGA High*/
	lgw_fpga_set_csn(1);

	/* determine return code */
	if (byte_transfered != ARRAY_SIZE(out_buf0)) {
		error("SPI WRITE FAILURE (byte_transfered = %d)", byte_transfered);
		return LGW_SPI_ERROR;
	} else {
		//debug("EEPROM write success");
		return LGW_SPI_SUCCESS;
	}
}

int lgw_spi_eeprom_cmd(uint8_t cmd)
{
	uint8_t out_buf0[3];
	uint8_t in_buf0[ARRAY_SIZE(out_buf0)];
	struct spi_ioc_transfer k;
	int byte_transfered = 0;

	/* prepare frame to be sent */
	out_buf0[0] = LGW_SPI_MUX_TARGET_EEPROM;
	out_buf0[1] = 0x01;
	out_buf0[2] = cmd;

	/* I/O transaction */
	memset(&k, 0, sizeof(k)); /* clear k */
	k.tx_buf = (unsigned long)out_buf0;
	k.rx_buf = (unsigned long)in_buf0;
	k.len = ARRAY_SIZE(out_buf0);
	k.speed_hz = EEPROM_SPI_SPEED;
	k.delay_usecs = 2;
	k.cs_change = 0;
	k.bits_per_word = 8;

	/*Set HOST_CSN_FPGA Low*/
	lgw_fpga_set_csn(0);
	/*Send SPI transaction*/
	byte_transfered = ioctl(spi_fpga_fd, SPI_IOC_MESSAGE(1), &k);

	/*Set HOST_CSN_FPGA High*/
	lgw_fpga_set_csn(1);

	/* determine return code */
	if (byte_transfered != ARRAY_SIZE(out_buf0)) {
		error("SPI WRITE FAILURE (byte_transfered = %d)", byte_transfered);
		return LGW_SPI_ERROR;
	} else {
		//debug("EEPROM cmd success");
		return LGW_SPI_SUCCESS;
	}
}
#endif

/* --- EOF ------------------------------------------------------------------ */
